import os
import sys
from xml.dom import minidom
import pmm_config
import pmm_dump_formatter
import pmm_api_xml_protocols
import subproc
import encodings.idna
from xml.parsers.expat import ExpatError

mswindows = (sys.platform == "win32")

pmm_ras_error_codes = {111: "Dump already exists in repository",
                       112: "Dump objects do not match with restore process owner",
                       121: "Permission denied error",
                       122: "FTP password does not match",
                       123: "FTP bad login",
                       124: "FTP host resolve error",
                       125: "FTP connection error",
                       126: "FTP network error",
                       127: "FTP dump file does not exists",
                       151: "Dump does not exist in repository",
                       152: "The file you are trying to upload is not a Plesk backup file,\
                             or it is a backup file from older version of Plesk. In the latter\
                             case you have to use the backup converter utility to convert this\
                             backup file into backup format supported by current version of Plesk.",
                       153: "Dump already exists"}

# Unknown error codes should be translated to error code = 199
pmm_ras_unknown_error_code = 199

class LocalRepositoryAccessService:
    def __init__(self, dumps_storage_credentials):
        self.__dumps_storage_credentials_formatter = pmm_dump_formatter.DumpsStorageCredentialsFormatter(dumps_storage_credentials)

    def _pmmras_get_dump_list(self, type, guid, id, name):
        pmmras_dump_list = None
        # get pmmras_dump_list as XML string from pmmras
        # call pmm_ras, passing type, guid, id, name
        password = self.__dumps_storage_credentials_formatter.get_password()
        os.environ['DUMP_STORAGE_PASSWD'] = str(password)
        cmd = subproc.CmdLine(pmm_config.pmm_ras())
        cmd.arg('--get-dump-list')
        if type:
            type_arg = '--type=' + type
            cmd.arg(type_arg)
        if guid:
            guid_arg = '--guid=' + guid
            cmd.arg(guid_arg)
        if id:
            id_arg = '--id=' + id
            cmd.arg(id_arg)
        if name:
            idna = name.find('xn--') >= 0
            name_arg = ''
            if idna:
                name_arg = '--name=' + name
            else:
                domain_names_parts = name.split('.')
                idna_domain_names_parts = []
                map((lambda part: idna_domain_names_parts.append(encodings.idna.ToASCII(part))), domain_names_parts)
                idna_name = '' 
                for part in idna_domain_names_parts:
                    if idna_name != '':
                        idna_name = idna_name + '.'
                    idna_name = idna_name + part
                name_arg = '--name=' + idna_name
            cmd.arg(name_arg)
        session_path_arg = '--session-path=' + pmm_config.pmm_logs_directory()
        cmd.arg(session_path_arg)
        errcode = 0
        message = u''
        try:
            proc = cmd.spawn()
            pmmras_dump_list = proc.stdout
        except subproc.NonzeroExitException, x:
            errcode = x.exitcode
            pmmras_dump_list = x.subprocess.stdout
            message =  u"pmm-ras error (Error code = " + unicode(errcode) + "):"  \
                     + u"\n== STDOUT ====================\n" + x.subprocess.stdout + u"\n==============================\n" \
                     + u"== STDERR ====================\n" + x.subprocess.stderr + u"\n==============================\n"

        dump_list = pmm_api_xml_protocols.DumpList.factory()
        if pmmras_dump_list:
            try:
                dump_list.build(minidom.parseString(pmmras_dump_list).childNodes[0])
            except ExpatError:
                message = message + u"Unable to parse DumpList: ExpatError"
        
        return dump_list, None, message

    def _pmmras_get_dump_status(self, dump_storage, dump_file):
        pmmras_dump_status = None
        # get pmmras_dump_status as XML string from pmmras
        # call pmm_ras, passing secure dump file specification
        password = self.__dumps_storage_credentials_formatter.get_password()
        os.environ['DUMP_STORAGE_PASSWD'] = str(password)
        cmd = subproc.CmdLine(pmm_config.pmm_ras())
        cmd.arg('--get-dump-info')
        dump_storage_arg = '--dump-storage=' + dump_storage
        cmd.arg(dump_storage_arg)
        dump_file_specification_arg = '--dump-file-specification=' + dump_file
        cmd.arg(dump_file_specification_arg)
        cmd.arg('--check-sign')
        session_path_arg = '--session-path=' + pmm_config.pmm_logs_directory()
        cmd.arg(session_path_arg)
        errcode = 0
        message = u''
        try:
            proc = cmd.spawn()
            pmmras_dump_status = proc.stdout
        except subproc.NonzeroExitException, x:
            errcode = x.exitcode
            if errcode in pmm_ras_error_codes.keys():
                message = pmm_ras_error_codes[errcode]
            else:
                # try get pmmras_dump_status from utility output
                pmmras_dump_status = x.subprocess.stdout
                message =  u"pmm-ras error (Error code = " + unicode(errcode) + "):"  \
                         + u"\n== STDOUT ====================\n" + x.subprocess.stdout + u"\n==============================\n" \
                         + u"== STDERR ====================\n" + x.subprocess.stderr + u"\n==============================\n"
                errcode = pmm_ras_unknown_error_code
        
        dump_result = None
        if pmmras_dump_status:
            dump_result = pmm_api_xml_protocols.Dump.factory()
            try:
                dump_result.build(minidom.parseString(pmmras_dump_status).childNodes[0])
            except ExpatError:
                message = message + u"Unable to parse Dump element: ExpatError"
        
        return dump_result, errcode, message

    def _pmmras_delete_dump(self, dump_storage, dump_file):
        # deletes dump
        # call pmm_ras, passing secure dump file specification
        password = self.__dumps_storage_credentials_formatter.get_password()
        os.environ['DUMP_STORAGE_PASSWD'] = str(password)
        cmd = subproc.CmdLine(pmm_config.pmm_ras())
        cmd.arg('--delete-dump')
        dump_storage_arg = '--dump-storage=' + dump_storage
        cmd.arg(dump_storage_arg)
        dump_file_specification_arg = '--dump-specification=' + dump_file
        cmd.arg(dump_file_specification_arg)
        session_path_arg = '--session-path=' + pmm_config.pmm_logs_directory()
        cmd.arg(session_path_arg)
        errcode = 0
        message = u''
        try:
            proc = cmd.spawn()
        except subproc.NonzeroExitException, x:
            errcode = x.exitcode
            if errcode in pmm_ras_error_codes.keys():
                message = pmm_ras_error_codes[errcode]
            else:
                message =  u"pmm-ras error (Error code = " + unicode(errcode) + "):"  \
                         + u"\n== STDOUT ====================\n" + x.subprocess.stdout + u"\n==============================\n" \
                         + u"== STDERR ====================\n" + x.subprocess.stderr + u"\n==============================\n"
                errcode = pmm_ras_unknown_error_code
        
        return errcode, message

    # the 'getDumpList' method returns XML-element-class instances autogenerated from 'pmm_api_xml_protocols.xsd'
    def getDumpList(self,object_specification):
        dump_list, errcode, message = self._pmmras_get_dump_list(object_specification.get_type(),object_specification.get_guid(),object_specification.get_id(),object_specification.get_name())
        return dump_list, errcode, message

    # the 'getDumpStatus' method returns XML-element-class instances autogenerated from 'pmm_api_xml_protocols.xsd'
    def getDumpStatus(self,dump_specification):
        dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(dump_specification)
        dump_file = dump_specification_formatter.get_name_of_xml_file()
        dump_storage = dump_specification_formatter.get_dumps_storage_credentials_formatter().getDumpStorage()
        dump_status, errcode, message = self._pmmras_get_dump_status(dump_storage, dump_file)
        return dump_status, errcode, message

    # the 'deleteDump' method returns error code and error message 
    def deleteDump(self,dump_specification):
        dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(dump_specification)
        dump_file = dump_specification_formatter.get_name_of_xml_file()
        dump_storage = dump_specification_formatter.get_dumps_storage_credentials_formatter().getDumpStorage()
        errcode, message = self._pmmras_delete_dump(dump_storage, dump_file)
        return errcode, message

class FtpRepositoryAccessService:
    def __init__(self, dumps_storage_credentials):
        self.__dumps_storage_credentials_formatter = pmm_dump_formatter.DumpsStorageCredentialsFormatter(dumps_storage_credentials)

    # for the case of migration functionality to PMM RAS
    def _pmmras_export_dump_as_file(self, dump_file_name, dump_storage_credentials_formatter, password):
        # dump_storage_credentials_formatter is a target DumpsStorageCredentialsFormatter
        if not mswindows:
            raise Exception("pmm-ras --export_dump_as_file is supported for Windows only!")
        
        os.environ['DUMP_STORAGE_PASSWD'] = str(password)
        cmd = subproc.CmdLine(pmm_config.pmm_ras())
        cmd.arg('--export-dump-as-file')
        
        if dump_storage_credentials_formatter.get_use_passive_ftp_mode() == 'true':
            cmd.arg('--use-ftp-passive-mode')
        
        # dump specification is a filename relative for local dumps storage
        dump_specification_arg = '--dump-specification=' + dump_file_name
        cmd.arg(dump_specification_arg)
        dump_file_specification_arg = '--dump-file-specification=' + dump_storage_credentials_formatter.getFileDumpStorage()
        cmd.arg(dump_file_specification_arg)
        session_path_arg = '--session-path=' + pmm_config.pmm_logs_directory()
        cmd.arg(session_path_arg)
        
        errcode = 0
        message = u''
        filename = None
        try:
            proc = cmd.spawn()
            filename = proc.stdout.strip()
        except subproc.NonzeroExitException, x:
            errcode = x.exitcode
            if errcode in pmm_ras_error_codes.keys():
                if errcode == 111:
                    filename = x.subprocess.stdout.strip()
                message = pmm_ras_error_codes[errcode]
            else:
                message =  u"pmm-ras error (Error code = " + unicode(errcode) + "):"  \
                         + u"\n== STDOUT ====================\n" + x.subprocess.stdout + u"\n==============================\n" \
                         + u"== STDERR ====================\n" + x.subprocess.stderr + u"\n==============================\n"
                errcode = pmm_ras_unknown_error_code
        
        return errcode, message, filename

    # download to local repository is supported only. so far pmm-ras does not need to know repository password
    def _pmmras_import_file_as_dump(self, dump_file, dumps_storage,guid,type):
        # call pmm_ras, passing secure_source_dump_file_specification,source_password,secure_dump_storage
        password = self.__dumps_storage_credentials_formatter.get_password()
        os.environ['DUMP_STORAGE_PASSWD'] = str(password)
        cmd = subproc.CmdLine(pmm_config.pmm_ras())
        cmd.arg('--import-file-as-dump')
        
        if self.__dumps_storage_credentials_formatter.get_use_passive_ftp_mode() == 'true':
            cmd.arg('--use-ftp-passive-mode')
        
        dump_file_specification_arg = '--dump-file-specification=' + dump_file
        cmd.arg(dump_file_specification_arg)
        dump_storage_arg = '--dump-storage=' + dumps_storage
        cmd.arg(dump_storage_arg)
        guid_arg = '--guid=' + guid
        cmd.arg(guid_arg)
        type_arg = '--type=' + type
        cmd.arg(type_arg)
        session_path_arg = '--session-path=' + pmm_config.pmm_logs_directory()
        cmd.arg(session_path_arg)
        errcode = 0
        message = u''
        filename = None
        try:
            proc = cmd.spawn()
            filename = proc.stdout.strip()
        except subproc.NonzeroExitException, x:
            errcode = x.exitcode
            if errcode in pmm_ras_error_codes.keys():
                if errcode == 111:
                    filename = x.subprocess.stdout.strip()
                message = pmm_ras_error_codes[errcode]
            else:
                message =  u"pmm-ras error (Error code = " + unicode(errcode) + "):"  \
                         + u"\n== STDOUT ====================\n" + x.subprocess.stdout + u"\n==============================\n" \
                         + u"== STDERR ====================\n" + x.subprocess.stderr + u"\n==============================\n"
                errcode = pmm_ras_unknown_error_code
        
        return errcode, message, filename

    def _pmmras_export_file_as_file(self, destination_dump, destination_dumps_storage_password):
        source_dump = self.__dumps_storage_credentials_formatter.getFileDumpStorage()
        source__dumps_storage_password = self.__dumps_storage_credentials_formatter.get_password()
        os.environ['DUMP_STORAGE_PASSWD_SRC'] = str(source__dumps_storage_password)
        os.environ['DUMP_STORAGE_PASSWD_DST'] = str(destination_dumps_storage_password)
        cmd = subproc.CmdLine(pmm_config.pmm_ras())
        cmd.arg('--export-file-as-file')
        
        if self.__dumps_storage_credentials_formatter.get_use_passive_ftp_mode() == 'true':
            cmd.arg('--use-ftp-passive-mode')
        
        file_from_arg = '--file-from=' + source_dump
        cmd.arg(file_from_arg)
        file_to_arg = '--file-to=' + destination_dump
        cmd.arg(file_to_arg)
        session_path_arg = '--session-path=' + pmm_config.pmm_logs_directory()
        cmd.arg(session_path_arg)
        errcode = 0
        message = u''
        try:
            proc = cmd.spawn()
        except subproc.NonzeroExitException, x:
            errcode = x.exitcode
            if errcode in pmm_ras_error_codes.keys():
                message = pmm_ras_error_codes[errcode]
            else:
                message =  u"pmm-ras error (Error code = " + unicode(errcode) + "):"  \
                         + u"\n== STDOUT ====================\n" + x.subprocess.stdout + u"\n==============================\n" \
                         + u"== STDERR ====================\n" + x.subprocess.stderr + u"\n==============================\n"
                errcode = pmm_ras_unknown_error_code
        
        return errcode, message

    def _pmmras_get_ftp_dump_list(self):
        password = self.__dumps_storage_credentials_formatter.get_password()
        secure_dumps_storage = self.__dumps_storage_credentials_formatter.getDumpStorage()
        os.environ['DUMP_STORAGE_PASSWD'] = str(password)
        pmmras_dump_list = None
        cmd = subproc.CmdLine(pmm_config.pmm_ras())
        cmd.arg('--get-ftp-dump-list')
        
        if self.__dumps_storage_credentials_formatter.get_use_passive_ftp_mode() == 'true':
            cmd.arg('--use-ftp-passive-mode')
        
        dump_storage_arg = '--dump-storage=' + secure_dumps_storage
        cmd.arg(dump_storage_arg)
        session_path_arg = '--session-path=' + pmm_config.pmm_logs_directory()
        cmd.arg(session_path_arg)
        errcode = 0
        message = u''
        try:
            proc = cmd.spawn()
            pmmras_dump_list = proc.stdout
        except subproc.NonzeroExitException, x:
            errcode = x.exitcode
            if errcode in pmm_ras_error_codes.keys():
                message = pmm_ras_error_codes[errcode]
            else:
                # try get pmmras_dump_list from utility output
                pmmras_dump_list = x.subprocess.stdout
                message =  u"pmm-ras error (Error code = " + unicode(errcode) + "):"  \
                         + u"\n== STDOUT ====================\n" + x.subprocess.stdout + u"\n==============================\n" \
                         + u"== STDERR ====================\n" + x.subprocess.stderr + u"\n==============================\n"
                errcode = pmm_ras_unknown_error_code
        
        dump_list = pmm_api_xml_protocols.DumpList.factory()
        if pmmras_dump_list:
            try:
                dump_list.build(minidom.parseString(pmmras_dump_list).childNodes[0])
            except ExpatError:
                message = message + u"\nUnable to parse DumpList: ExpatError"
        
        return dump_list, None, message

    def _pmmras_get_dump_status(self,dump_storage,dump_file):
        # get pmmras_dump_status as XML string from pmmras
        # call pmm_ras, passing secure dump file specification
        password = self.__dumps_storage_credentials_formatter.get_password()
        os.environ['DUMP_STORAGE_PASSWD'] = str(password)
        pmmras_dump_status = None
        cmd = subproc.CmdLine(pmm_config.pmm_ras())
        cmd.arg('--get-ftp-dump-info')
        
        if self.__dumps_storage_credentials_formatter.get_use_passive_ftp_mode() == 'true':
            cmd.arg('--use-ftp-passive-mode')
        
        dump_storage_arg = '--dump-storage=' + dump_storage
        cmd.arg(dump_storage_arg)
        dump_file_specification_arg = '--dump-file-specification=' + dump_file
        cmd.arg(dump_file_specification_arg)
        session_path_arg = '--session-path=' + pmm_config.pmm_logs_directory()
        cmd.arg(session_path_arg)
        errcode = 0
        message = u''
        try:
            proc = cmd.spawn()
            pmmras_dump_status = proc.stdout
        except subproc.NonzeroExitException, x:
            errcode = x.exitcode
            if errcode in pmm_ras_error_codes.keys():
                message = pmm_ras_error_codes[errcode]
            else:
                # try get pmmras_dump_status from utility output
                pmmras_dump_status = x.subprocess.stdout
                message =  u"pmm-ras error (Error code = " + unicode(errcode) + "):"  \
                         + u"\n== STDOUT ====================\n" + x.subprocess.stdout + u"\n==============================\n" \
                         + u"== STDERR ====================\n" + x.subprocess.stderr + u"\n==============================\n"
                errcode = pmm_ras_unknown_error_code
        
        dump_result = None
        if pmmras_dump_status:
            dump_result = pmm_api_xml_protocols.Dump.factory()
            try:
                dump_result.build(minidom.parseString(pmmras_dump_status).childNodes[0])
            except ExpatError:
                message = message + u"\nUnable to parse Dump element: ExpatError"
        
        return dump_result, errcode, message

    def _pmmras_delete_dump(self, dump_storage, dump_file):
        # deletes dump
        password = self.__dumps_storage_credentials_formatter.get_password()
        os.environ['DUMP_STORAGE_PASSWD'] = str(password)
        cmd = subproc.CmdLine(pmm_config.pmm_ras())
        cmd.arg('--delete-exported-dump')
        
        if self.__dumps_storage_credentials_formatter.get_use_passive_ftp_mode() == 'true':
            cmd.arg('--use-ftp-passive-mode')
        
        dump_storage_arg = '--dump-storage=' + dump_storage
        cmd.arg(dump_storage_arg)
        dump_file_specification_arg = '--dump-file-specification=' + dump_file
        cmd.arg(dump_file_specification_arg)
        session_path_arg = '--session-path=' + pmm_config.pmm_logs_directory()
        cmd.arg(session_path_arg)
        errcode = 0
        message = u''
        try:
            proc = cmd.spawn()
        except subproc.NonzeroExitException, x:
            errcode = x.exitcode
            if errcode in pmm_ras_error_codes.keys():
                message = pmm_ras_error_codes[errcode]
            else:
                message =  u"pmm-ras error (Error code = " + unicode(errcode) + "):"  \
                         + u"\n== STDOUT ====================\n" + x.subprocess.stdout + u"\n==============================\n" \
                         + u"== STDERR ====================\n" + x.subprocess.stderr + u"\n==============================\n"
                errcode = pmm_ras_unknown_error_code
        
        return errcode, message

    # the 'import_file_as_dump' method returns errcode, message
    # pmm-ras supports only 'local' storage-type in destination_dump_specification
    def import_file_as_dump(self,source_dump_specification,destination_dump_specification,guid,type):
        source_dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(source_dump_specification)
        destination_dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(destination_dump_specification)
        dump_file = source_dump_specification_formatter.get_destination_file()
        dumps_storage = destination_dump_specification_formatter.get_dumps_storage_credentials_formatter().getDumpStorage()
        errcode, message, filename = self._pmmras_import_file_as_dump(dump_file,dumps_storage,guid,type)
        return errcode, message, filename
        
    # the 'export_dump_as_file' method returns errcode, message
    def export_dump_as_file(self,source_dump_specification,destination_dump_specification):
        source_dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(source_dump_specification)
        dump_file_name = source_dump_specification_formatter.get_name_of_xml_file()
        destination_dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(destination_dump_specification)
        dumps_storage_credentials_formatter = destination_dump_specification_formatter.get_dumps_storage_credentials_formatter()
        destination_dumps_storage_password = dumps_storage_credentials_formatter.get_password()
        errcode, message, filename = self._pmmras_export_dump_as_file(dump_file_name,dumps_storage_credentials_formatter,destination_dumps_storage_password)
        return errcode, message, filename

    # the 'export_file_as_file' method returns errcode, message
    def export_file_as_file(self,destination_dump_specification):
        destination_dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(destination_dump_specification)
        destination_dump = destination_dump_specification_formatter.get_destination_file()
        destination_dumps_storage_password = destination_dump_specification_formatter.get_dumps_storage_credentials_formatter().get_password()
        errcode, message = self._pmmras_export_file_as_file(destination_dump, destination_dumps_storage_password)
        return errcode, message

    # the 'getDumpList' method returns XML-element-class instances autogenerated from 'pmm_api_xml_protocols.xsd'
    def getDumpList(self,object_specification):
        # object specification is ignored for ftp dumps, pmm-ras shows all dump files
        dump_list, errcode, message = self._pmmras_get_ftp_dump_list()
        return dump_list, errcode, message

    # the 'getDumpStatus' method returns XML-element-class instances autogenerated from 'pmm_api_xml_protocols.xsd'
    def getDumpStatus(self,dump_specification):
        dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(dump_specification)
        dump_storage_credentials_formatter = dump_specification_formatter.get_dumps_storage_credentials_formatter()
        dump_storage = dump_storage_credentials_formatter.getDumpStorage()
        dump_file = dump_storage_credentials_formatter.get_file_name()
        dump_status, errcode, message = self._pmmras_get_dump_status(dump_storage,dump_file)
        return dump_status, errcode, message

    # the 'deleteDump' method returns error code and error message 
    def deleteDump(self,dump_specification):
        dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(dump_specification)
        dump_storage_credentials_formatter = dump_specification_formatter.get_dumps_storage_credentials_formatter()
        dump_file = dump_storage_credentials_formatter.get_file_name()
        dump_storage = dump_storage_credentials_formatter.getDumpStorage()
        errcode, message = self._pmmras_delete_dump(dump_storage, dump_file)
        return errcode, message

