from threading import RLock, currentThread
import os, logging
import shutil
import osutil
import signal
import pmm_config
import pmm_dump_formatter
import pmmcli_daemon_service
import pmm_dump_access_service
import validator
from xml.dom import minidom
from xml.parsers.expat import ExpatError
from pmm_api_xml_protocols import TaskStatus, Stopped, Finished, TaskLog, WorkingProgress, Working, Starting, TaskList, Task
from execution_result import ExecutionResult
import marshal
import time


_logger = logging.getLogger("pmmcli.task")


class DumpMarshaller:
    def __init__(self,storage):
        self.storage = storage

    def dump(self,data):
        raise Exception, "Could not invoke abstract method 'dump' of " + str(self.__class__)

    def close(self):
        raise Exception, "Could not invoke abstract method 'close' of " + str(self.__class__)


class LoadMarshaller:
    def __init__(self,storage):
        self.storage = storage

    def load(self):
        raise Exception, "Could not invoke abstract method 'load' of " + str(self.__class__)

    def close(self):
        raise Exception, "Could not invoke abstract method 'close' of " + str(self.__class__)


class FileDumpMarshaller(DumpMarshaller):
    def __init__(self,filename):
        DumpMarshaller.__init__(self,open(filename,'wb'))
        self.__filename = filename
        _logger.debug("initialized dump marshaller "+self.__filename)

    def dump(self, data):
        bytes = marshal.dumps(data)
        self.storage.write(bytes)

    def close(self):
        self.storage.close()


class FileLoadMarshaller(LoadMarshaller):
    def __init__(self,filename):
        LoadMarshaller.__init__(self,open(filename,'rb'))
        self.__filename = filename
        _logger.debug("initialized load marshaller "+self.__filename)

    def load(self):
        loadeddata = marshal.load(self.storage)
        return loadeddata

    def close(self):
        self.storage.close()

# all and only the objects of types 'int' and 'str' are serialized
class PmmTask:
    def __init__(self, cmd, pid, owner_guid, owner_type, session_path, dumps_storage_credentials, fullname = "", name = "", tasktype = "None", backup_profile_name = "", mailto = "", topobject_id = None, topobject_type = None):
        self.__tasktype = str(tasktype)
        self.__cmd = str(cmd)
        self.__os_pid = int(pid)
        self.__pmm_pid = int(pid)
        self.__owner_guid = str(owner_guid)
        self.__owner_type = str(owner_type)
        # dumps_storage_credentials has string format
        self.__dumps_storage_credentials = str(dumps_storage_credentials)
        self.__creation_date = time.time()
        self.__fullname = str(fullname)
        self.__name = str(name)
        self.__backup_profile_name = str(backup_profile_name)
        self.__mailto = str(mailto)
        self.__topobject_id = str(topobject_id)
        self.__topobject_type = str(topobject_type)
        
        self.__mailsent = 0
        
        self.__host_credentials = ""
        self.__log_filename = ""
        if tasktype == 'Backup':
            self.__log_filename = os.path.join(session_path,'psadump.log')
        elif tasktype == 'Restore':
            self.__log_filename = os.path.join(session_path,'migration.log')
        self.__migration_result_filename = os.path.join(session_path,'migration.result')
        self.__session_path = str(session_path)
        self.__stateMutex = RLock()
        self.__stopped = 0

    def is_active(self):
        active = osutil.is_active(self.__os_pid)
        if active:
            process_cmd = osutil.get_cmd(self.__os_pid)
            # This deal with 'get_cmd' implementation differences on Windows and Linux platforms
            same_cmd = self.__cmd.startswith(process_cmd)
            if not same_cmd:
                _logger.debug("process (%s) found with the same pid as %s task pid (%d)" % (process_cmd,self.__tasktype,self.__os_pid))
        return active and same_cmd

    def dump(self, dump_marshaller):
        pdump = dump_marshaller.dump
        key_list = self.__dict__.keys()
        key_list.sort()
        for key in key_list:
            key_type = type(self.__dict__[key])
            if key_type in [str,int,float]:
                pdump(key)
                pdump(self.__dict__[key])

    def load(self, load_marshaller):
        pload = load_marshaller.load
        key_list = self.__dict__.keys()
        key_list.sort()
        for key in key_list:
            key_type = type(self.__dict__[key])
            if key_type in [str,int,float]:
                key_name = ""
                key_name = pload()
                self.__dict__[key] = pload()

    def get_os_pid(self):
        return self.__os_pid

    def get_pmm_pid(self):
        return self.__pmm_pid

    def get_owner_guid(self):
        return self.__owner_guid

    def get_owner_type(self):
        return self.__owner_type

    def get_dumps_storage_credentials(self):
        return self.__dumps_storage_credentials

    def get_tasktype(self):
        return self.__tasktype

    def get_host_credentials(self):
        return self.__host_credentials

    def get_log_filename(self):
        return self.__log_filename

    def get_migration_result_filename(self):
        return self.__migration_result_filename

    def get_session_path(self):
        return self.__session_path

    def get_fullname(self):
        return self.__fullname

    def get_creation_date_formatted(self):
        return time.strftime('%Y-%b-%d %H:%M:%S',time.localtime(self.__creation_date))

    def get_creation_date(self):
        return self.__creation_date

    def get_name(self):
        return self.__name

    def get_backup_profile_name(self):
        return self.__backup_profile_name

    def get_mailto(self):
        return self.__mailto

    def get_mailsent(self):
        return self.__mailsent != 0

    def get_topobject_id(self):
        return self.__topobject_id

    def get_topobject_type(self):
        return self.__topobject_type

    def get_stopped(self):
        return self.__stopped

    def set_os_pid(self, os_pid):
        self.__os_pid = os_pid

    def set_pmm_pid(self, pmm_pid):
        self.__pmm_pid = pmm_pid

    def set_tasktype(self, tasktype):
        self.__tasktype = tasktype

    def set_host_credentials(self, host_credentials):
        self.__host_credentials = host_credentials

    def set_log_filename(self, log_filename):
        self.__log_filename = str(log_filename)

    def set_session_path(self, session_path):
        self.__session_path = session_path

    def set_name(self, name):
        self.__name = str(name)

    def set_fullname(self, fullname):
        self.__fullname = str(fullname)

    def set_mailsent(self):
        self.__mailsent = 1

    def set_owner_guid(self, owner_guid):
        self.__owner_guid = owner_guid

    def set_owner_type(self, owner_type):
        self.__owner_type = owner_type

    def set_dumps_storage_credentials(self, set_dumps_storage_credentials):
        self.__dumps_storage_credentials = set_dumps_storage_credentials

    def set_stopped(self):
        self.__stopped = 1


class BackupTask(PmmTask):
    def __init__(self, cmd, pid, owner_guid, owner_type, session_path, dumps_storage_credentials, backup_profile_name, mailto, topobject_id, topobject_type):
        PmmTask.__init__(self, cmd, pid, owner_guid, owner_type, session_path, dumps_storage_credentials, "", "", "Backup", backup_profile_name, mailto, topobject_id, topobject_type )
        getPMMTaskManager().addTask(self)


class RestoreTask(PmmTask):
    def __init__(self, cmd, pid, owner_guid, owner_type, session_path, dumps_storage_credentials, full_name, name, mailto):
        PmmTask.__init__(self, cmd, pid, owner_guid, owner_type, session_path, dumps_storage_credentials, full_name, name, "Restore", "", mailto, None, None)
        getPMMTaskManager().addTask(self)


execution_result_validator = None


def getExecutionResultValidator():
    global execution_result_validator
    if not execution_result_validator:
        execution_result_validator = XsdValidator(pmm_config.execution_result_schema())
    return execution_result_validator


class XsdValidator:
    def __init__(self, schema):
        self.__validator = validator.Validator(schema)

    def do(self, xml):
        return self.__validator.do(xml)


class PMMTaskPersistentStorage(object):
    def __init__(self):
        self.accessMutex = RLock()
        self.persistent_storage_dir = pmm_config.tasks_dir()

    def getPmmTask(self,task_id):
        task_id_str = str(task_id)
        tasksDirectory = pmm_config.tasks_dir()
        if os.path.isdir(tasksDirectory):
            fullname = os.path.join(tasksDirectory, task_id_str)
            if os.path.isfile(fullname):
                loader = FileLoadMarshaller(fullname)
                task_item = PmmTask("",0,"0000-0000-0000-0000","","","")
                try:
                    task_item.load(loader)
                    loader.close()
                except (ValueError, EOFError), e:
                    _logger.error("Could not load task (%s) data: %s" % (task_id_str,str(e)))
                    return None
                return task_item

    def getPmmTasks(self,types,owner_guid,topobject_id,topobject_type):
        tasksDirectory = pmm_config.tasks_dir()
        task_list = []
        if os.path.isdir(tasksDirectory):
            names = os.listdir(tasksDirectory)
            for name in names:
                fullname = os.path.join(tasksDirectory, name)
                if os.path.isfile(fullname):
                    # name is a task number
                    loader = FileLoadMarshaller(fullname)
                    task_item = PmmTask("",0,"0000-0000-0000-0000","","","")
                    try:
                        task_item.load(loader)
                        loader.close()
                    except (ValueError, EOFError), e:
                        _logger.error("Could not load task (%s) data:\n %s" % (name,str(e)))
                        continue
                    if task_item.get_tasktype() in types:
                        # FS specifies that owner_guid, topobject_id, topobject_type are satisfied with followings:
                        # if some object is None, the rest objects in list are None too
                        if owner_guid:
                            if not owner_guid in ["any", task_item.get_owner_guid()]:
                                continue
                            elif topobject_id:
                                if not topobject_id in ["any", task_item.get_topobject_id()]:
                                    continue
                                elif topobject_type:
                                    if not topobject_type == task_item.get_topobject_type():
                                        continue
                        task_list.append(task_item)
        # sort task list by creation_date
        temp_array = [(-item.get_creation_date(), index, item) for index, item in enumerate(task_list)]
        temp_array.sort()
        return [item[-1] for item in temp_array]

    def getUniquePmmPid(self):
        tasksDirectory = pmm_config.tasks_dir()
        pmmPids = []
        if os.path.isdir(tasksDirectory):
            pmmPids = os.listdir(tasksDirectory)
        uniquePmmPid = -1
        while str(uniquePmmPid) in pmmPids:
            uniquePmmPid = uniquePmmPid - 1
        return uniquePmmPid

    def addPmmTask(self, pmmTask):
        if not os.path.isdir(self.persistent_storage_dir):
            os.mkdir(self.persistent_storage_dir)
        dumper = FileDumpMarshaller(os.path.join(self.persistent_storage_dir,str(pmmTask.get_pmm_pid())))
        pmmTask.dump(dumper)
        dumper.close()

    def stopPmmTask(self, pmmTask):
        if not os.path.isdir(self.persistent_storage_dir):
            os.mkdir(self.persistent_storage_dir)
        pmmTask.set_stopped()
        dumper = FileDumpMarshaller(os.path.join(self.persistent_storage_dir,str(pmmTask.get_pmm_pid())))
        pmmTask.dump(dumper)
        dumper.close()

    def removePmmTask(self, pmmTask):
        osutil.unlink_recursively(pmmTask.get_session_path())
        osutil.unlink_nothrow(os.path.join(self.persistent_storage_dir,str(pmmTask.get_pmm_pid())))


class FinishedTaskStatusCreator:
    def __init__(self, task):
        self.__task = task
        if self.__task.get_fullname() == '':
            dump_file_name_file_name = os.path.join(self.__task.get_session_path(),'dump-name')
            if os.path.isfile(dump_file_name_file_name):
                dump_file_name_file = open(dump_file_name_file_name,"rt")
                # pmm-ras for local storage on unix returns file name that starts with '/'
                # we should remove it from filename
                dump_file_name = dump_file_name_file.readline().rstrip('\n').lstrip('/\\')
                dump_file_name_file.close()
                if dump_file_name != '':
                    # dump file name here could be
                    formatter = pmm_dump_formatter.DumpsStorageCredentialsFormatter(self.__task.get_dumps_storage_credentials())
                    storage = formatter.getDumpStorageNoLogin()
                    if formatter.get_storage_type() == 'foreign-ftp':
                        # pmm-ras could return mixed separators in the name of exported file for ftp storage type
                        dump_file_name = dump_file_name.replace('\\','/')
                    file_name = dump_file_name.replace(storage,'')
                    self.__task.set_fullname(file_name)
                    self.__task.set_name(file_name)
                    getPMMTaskManager().updateTask(self.__task)
        err_file_name = os.path.join(self.__task.get_session_path(),".stderr")
        err_string = None
        if os.path.isfile(err_file_name):
            err = open(err_file_name,"r")
            err_string = err.read()
            err.close()
            shutil.copy2(err_file_name,os.path.join(self.__task.get_session_path(),"stderr"))
            osutil.unlink_nothrow(err_file_name)
        
        if err_string and err_string != '':
            stderr_string = "\n== STDERR ====================\n" + err_string + "\n==============================\n"
            log_file_name = self.__task.get_log_filename()
            if os.path.isfile(log_file_name):
                log_file = open(log_file_name,"a")
                log_file.write(stderr_string)
                log_file.close()

    def response(self):
        task_status = None
        pid = str(self.__task.get_pmm_pid())
        status_file_path = os.path.join(self.__task.get_session_path(),'migration.result')
        _logger.debug("Get task (%s) status from migration.result..." % pid)
        _logger.debug("Session path (%s)" % self.__task.get_session_path())
        if os.path.isfile(status_file_path):
            #load migration result from file
            try:
                getter = pmm_dump_access_service.XmlElementTextGetter('execution-result')
                status_getter = pmm_dump_access_service.XmlElementAttrGetter('execution-result','status')
                xml_processor = pmm_dump_access_service.InfoXMLProcessor( status_file_path, pmm_dump_access_service.XmlHandler([getter,status_getter]))
                xml_processor.process()
                status_attribute = status_getter.get_attribute_value()
                # task working process (backup or deployer) writes only 'error' or 'success' as status of finished task
                # if status task is not 'error' we should suppose it is 'success'
                if status_attribute != 'error':
                    status_attribute = 'success'
                status_value = getter.get_element_inner_text()
                # if execution result has any context then reduce status to 'warning'
                if status_attribute == 'success' and status_value.strip() != '':
                    status_attribute = 'warnings'
                task_status = Finished( status = status_attribute, log_location = self.__task.get_migration_result_filename())
            except:
                _logger.debug("Execution result file %s exists but has wrong format. The task %s will be marked as finished with success." % (status_file_path, pid))
                task_status = Finished( status = 'success', log_location = self.__task.get_migration_result_filename())
        else:
            _logger.debug("Execution result file %s isn't found. The task %s will be marked as finished with error." % (status_file_path, pid))
            task_status = Finished( status = 'error', log_location = self.__task.get_migration_result_filename())

        return TaskStatus ( finished = task_status )


class StoppedTaskStatusCreator:
    def __init__(self, task):
        self.__task = task

    def response(self):
        return TaskStatus ( stopped = Stopped() )


class TaskLogCreator:
    def __init__(self, task_log):
        self.__task_log = task_log

    def response(self):
        return TaskLog ( self.__task_log )


class RunningTaskStatusCreator:
    def __init__(self, task):
        self.__task = task

    def response(self):
        task_status = None
        if self.__task.get_tasktype() == 'Restore':
            status_file_path = os.path.join(self.__task.get_session_path(),'migration.status')
            if os.path.isfile(status_file_path):
                #load deploy status from file
                status_file = open(status_file_path,"rt")
                status_file_content = status_file.read()
                status_file.close()
                deploy_status = WorkingProgress.factory()
                deploy_status.build(minidom.parseString(status_file_content).childNodes[0])
                task_status = Working( deploy = deploy_status )
            else:
                task_status = Working( starting = Starting() )
        else:
            status_file_path = os.path.join(self.__task.get_session_path(),'dump-status.xml')
            if os.path.isfile(status_file_path):
                #load dumping status from file
                status_file = open(status_file_path,"rt")
                status_file_content = status_file.read()
                status_file.close()
                dumping_status = WorkingProgress.factory()
                dumping_status.build(minidom.parseString(status_file_content).childNodes[0])
                task_status = Working( dumping = dumping_status )
            else:
                task_status = Working( starting = Starting() )
        return TaskStatus ( working = task_status )

pmmTaskManager = None


def getPMMTaskManager():
    global pmmTaskManager
    if not pmmTaskManager:
        pmmTaskManager = PMMTaskManager()
    return pmmTaskManager


class PMMTaskManager(object):
    def __init__(self):
        self.persistent_storage = PMMTaskPersistentStorage()
        self.stateMutex = RLock()

    def realGetTaskList(self, types, owner_guid = None, topobject_id = None, topobject_type = None):
        task_list = self.persistent_storage.getPmmTasks(types,owner_guid,topobject_id,topobject_type)
        for task in task_list:
            self.__operatorUpdateTaskStatus(task)
        return task_list

    def __operatorGetTaskLog(self,task):
        # read task log and return as string
        log_file = task.get_log_filename()
        log_string = "Log file \'" + log_file + "\' does not exists"
        if os.path.isfile(log_file):
            log = open(log_file,"rt")
            log_string = log.read().decode('latin-1')
            log.close()
        
        return TaskLogCreator(log_string.encode('utf-8')).response()

    def __operatorUpdateTaskStatus(self, task):
        if not task.get_stopped():
            if not osutil.is_active(task.get_os_pid()):
                FinishedTaskStatusCreator(task)

    def operatorGetTaskStatus(self, task):
        if task.get_stopped():
            return StoppedTaskStatusCreator(task).response()
        else:
            active = task.is_active()
            if active:
                return RunningTaskStatusCreator(task).response()
            else:
                return FinishedTaskStatusCreator(task).response()

    def __operatorGetTask(self, task):
        task_list = TaskList.factory()
        dumps_storage_credentials = task.get_dumps_storage_credentials()
        if dumps_storage_credentials:
            dumps_storage_credentials_formatter = pmm_dump_formatter.DumpsStorageCredentialsFormatter(dumps_storage_credentials)
            task = Task.factory(task_id = task.get_pmm_pid(), owner_guid = task.get_owner_guid(), owner_type = task.get_owner_type(), fullname = task.get_fullname(), creation_date = task.get_creation_date_formatted(), name = task.get_name(), backup_profile_name = task.get_backup_profile_name(), mail_to = task.get_mailto(), task_status = self.operatorGetTaskStatus(task), dumps_storage_credentials = dumps_storage_credentials_formatter.buildXml())
        else :
            task = Task.factory(task_id = task.get_pmm_pid(), owner_guid = task.get_owner_guid(), owner_type = task.get_owner_type(), fullname = task.get_fullname(), creation_date = task.get_creation_date_formatted(), name = task.get_name(), backup_profile_name = task.get_backup_profile_name(), mail_to = task.get_mailto(), task_status = self.operatorGetTaskStatus(task))
        task_list.add_task(task)
        return task_list

    def __operatorStopTask(self, task):
        if not task.get_stopped():
            try:
                if task.is_active():
                    osutil.kill(task.get_os_pid(), signal.SIGTERM)
            except OSError, e:
                _logger.warning("__operatorStopTask failed: "+ str(e))
            self.persistent_storage.stopPmmTask(task)
        return 1

    def __operatorRemoveTask(self,task):
        self.__operatorStopTask(task)
        self.persistent_storage.removePmmTask(task)
        return 1

    def __taskFinder(self, task_id, operator):
        task = self.persistent_storage.getPmmTask(task_id)
        if task:
            self.__operatorUpdateTaskStatus(task)
            return operator(task)

    def updateTask(self, pmmTask):
        _logger.debug("Acquired session mutex: " + currentThread().getName())
        self.stateMutex.acquire()
        try:
            self.persistent_storage.addPmmTask(pmmTask)
        finally:
            self.stateMutex.release()
            _logger.debug("Released session mutex: " + currentThread().getName())

    def addTask(self, pmmTask):
        _logger.debug("Acquired session mutex: " + currentThread().getName())
        self.stateMutex.acquire()
        try:
            if pmmTask.get_pmm_pid() < 0:
                pmmTask.set_stopped()
                pmmTask.set_pmm_pid(self.persistent_storage.getUniquePmmPid())
            else:
                # start pmmcli_daemon
                pmmcli_daemon_service.PMMCliDaemon().start()
            self.persistent_storage.addPmmTask(pmmTask)
        
        finally:
            self.stateMutex.release()
            _logger.debug("Released session mutex: " + currentThread().getName())

    def stopTask(self, task_id):
        _logger.debug("Acquired session mutex: " + currentThread().getName())
        self.stateMutex.acquire()
        try:
            result = self.__taskFinder(task_id,self.__operatorStopTask)
        finally:
            self.stateMutex.release()
            _logger.debug("Released session mutex: " + currentThread().getName())
        return result		

    def removeTask(self, task_id):
        _logger.debug("Acquired session mutex: " + currentThread().getName())
        self.stateMutex.acquire()
        try:
            result = self.__taskFinder(task_id,self.__operatorRemoveTask)
        finally:
            self.stateMutex.release()
            _logger.debug("Released session mutex: " + currentThread().getName())
        return result		

    # the 'getXxxx' methods return XML-element-class instances autogenerated from 'pmm_api_xml_protocols.xsd'
    def getTaskList(self, types, owner_guid, topobject_id, topobject_type):
        pmm_task_list = []
        _logger.debug("Acquired session mutex: " + currentThread().getName())
        self.stateMutex.acquire()
        try:
            pmm_task_list = self.realGetTaskList(types,owner_guid,topobject_id,topobject_type)
        finally:
            self.stateMutex.release()
            _logger.debug("Released session mutex: " + currentThread().getName())
        task_list = TaskList.factory()
        for task_item in pmm_task_list:
            task = None
            dumps_storage_credentials = task_item.get_dumps_storage_credentials()
            if dumps_storage_credentials:
                dumps_storage_credentials_formatter = pmm_dump_formatter.DumpsStorageCredentialsFormatter(dumps_storage_credentials)
                task = Task.factory(task_id = task_item.get_pmm_pid(), owner_guid = task_item.get_owner_guid(), owner_type = task_item.get_owner_type(), fullname = task_item.get_fullname(), creation_date = task_item.get_creation_date_formatted(), name = task_item.get_name(), backup_profile_name = task_item.get_backup_profile_name(), mail_to = task_item.get_mailto(), task_status = self.operatorGetTaskStatus(task_item), dumps_storage_credentials = dumps_storage_credentials_formatter.buildXml())
            else :
                task = Task.factory(task_id = task_item.get_pmm_pid(), owner_guid = task_item.get_owner_guid(), owner_type = task_item.get_owner_type(), fullname = task_item.get_fullname(), creation_date = task_item.get_creation_date_formatted(), name = task_item.get_name(), backup_profile_name = task_item.get_backup_profile_name(), mail_to = task_item.get_mailto(), task_status = self.operatorGetTaskStatus(task_item))
            task_list.add_task(task)
        return task_list

    def getTask(self, task_id):
        _logger.debug("Acquired session mutex: " + currentThread().getName())
        self.stateMutex.acquire()
        try:
            result = self.__taskFinder(task_id,self.__operatorGetTask)
        finally:
            self.stateMutex.release()
            _logger.debug("Released session mutex: " + currentThread().getName())
        return result

    def getTaskStatus(self, task_id):
        _logger.debug("Acquired session mutex: " + currentThread().getName())
        self.stateMutex.acquire()
        try:
            result = self.__taskFinder(task_id,self.operatorGetTaskStatus)
        finally:
            self.stateMutex.release()
            _logger.debug("Released session mutex: " + currentThread().getName())
        return result		

    def getTaskLog(self, task_id):
        _logger.debug("Acquired session mutex: " + currentThread().getName())
        self.stateMutex.acquire()
        try:
            result = self.__taskFinder(task_id,self.__operatorGetTaskLog)
        finally:
            self.stateMutex.release()
            _logger.debug("Released session mutex: " + currentThread().getName())
        return result

