import os
import log
import time
import random
import libxml2
import stacktrace
import cStringIO, logging
import sys
from xml.dom import minidom
from xml.parsers.expat import ExpatError

import pmm_config
import plesk_config
import pmmcli_config
import pmm_task
import pmmcli_session
import pmmcli_daemon_service
from pmmcli_daemon_actions import DaemonTaskDeleteDump
import pmm_repository_access_service
import pmm_dump_formatter
from pmm_conflict_detector import PMMUtilityException, PleskActualDump
from pmm_api_xml_protocols import DumpListQuery, Data, RestoreTaskDescription, RestoreTaskResult, DumpValidityError, TaskId, BackupTaskDescription, ResolveConflictsTaskDescription, Response, ConflictResolutionRules, SrcDstFilesSpecification, CheckDumpResult
import osutil
import subproc
import validator
from encoder import EncoderFile

mswindows = (sys.platform == "win32")

trace = True

_logger = logging.getLogger("pmm.pmmcli")


libxml2errorHandlerErr = ''
def libxml2errorHandler(ctx, str):
    global libxml2errorHandlerErr
    libxml2errorHandlerErr = libxml2errorHandlerErr + "%s %s" % (ctx, str)


class PmmcliException(Exception):
    def __init__(self):
        self.message = None


class PmmcliActionParamException(PmmcliException):
    def __init__(self, PmmcliAction, message):
        self.caller = PmmcliAction
        self.msg = message

    def get_message(self):
        return self.msg


class ActionRunner(object):
    def __init__(self, ActionProcessor,parameter_stdin = None, parameters = None):
        self.processor = ActionProcessor(parameter_stdin, parameters)

    def doActivity(self):
        libxml2.registerErrorHandler(libxml2errorHandler, '')
        input_valid, msg = self.processor.validate()
        if not input_valid:
            _logger.error("Validate failed: ActionProcessor " + str(self.processor))
            raise PmmcliActionParamException(self.processor, msg)
        result = None
        _logger.debug(str(self) + ": doActivity")
        result = self.processor.doActivity()
        return self.processor.response(result)


class PMMCliAction(object):
    def __init__(self, parameter_stdin, parameters):
        _logger.debug("--> " + str(self))
        if parameters:
            _logger.info("parameters: " + str(parameters))
        if parameter_stdin:
            _logger.info("stdin: " + parameter_stdin)
        self._stdin = parameter_stdin
        self._params = parameters
        self._process = None
        self.__input_validator = validator.Validator(pmm_config.pmm_api_xml_protocols_schema())

    def get_input_validator(self):     
        return self.__input_validator

    def validate(self):
        _logger.debug(str(self) + ": validate")
        return 1

    def doActivity(self):
        raise PmmcliException()

    def response(self, result):
        _logger.debug(str(self) + ": response")
        return result

    def setProcess(self, process):
        if process == None:
            _logger.debug("Subprocess finished")
        else:
            _logger.debug(u"Subprocess " + unicode(process) + u" encountered.")
            self._process = process


class GetDumpsListAction(PMMCliAction):
    def __init__(self, parameter_stdin, parameters):
        PMMCliAction.__init__(self, parameter_stdin, parameters)
        self.__parameter_stdin = parameter_stdin
        self.__dump_list_query = None

    def validate(self):
        try:
            error_code, msg = self.get_input_validator().do(self.__parameter_stdin)
        except libxml2.parserError, ex:
            return None, 'XML parse error: ' + ex.msg + '\n' + libxml2errorHandlerErr
        if error_code:
            return None, 'Error ' + str(error_code) + ': ' + msg
        try:
            self.__dump_list_query = DumpListQuery.factory()
            self.__dump_list_query.build(minidom.parseString(self.__parameter_stdin).childNodes[0])
        except ExpatError:
            return None, 'ExpatError in DumpListQuery'
        return 1, ''

    def doActivity(self):
        if not self.__dump_list_query:
            self.__dump_list_query = DumpListQuery.factory()
            self.__dump_list_query.build(minidom.parseString(self.__parameter_stdin).childNodes[0])
        dumps_storage_credentials = self.__dump_list_query.get_dumps_storage_credentials()
        object_specification =  self.__dump_list_query.get_object_specification()
        storage_type = dumps_storage_credentials.get_storage_type()
        access_service = None
        if storage_type != 'local':
            access_service = pmm_repository_access_service.FtpRepositoryAccessService
        else:
            access_service = pmm_repository_access_service.LocalRepositoryAccessService
        result_dump_list, errcode, message = access_service(dumps_storage_credentials).getDumpList(object_specification)
        data = Data.factory( dump_list = result_dump_list )
        return data, errcode, message


class GetDumpOverviewAction(PMMCliAction):
    def __init__(self, parameter_stdin, parameters):
        PMMCliAction.__init__(self, parameter_stdin, parameters)
        self.__parameter_stdin = parameter_stdin
        self.__session_id = None
        self.__dump_specification_formatter = None

    def validate(self):
        # there is no validation - if stdin is not DumpSpecification suppose it is session-id
        return 1, ''

    def doActivity(self):
        session = None
        try:
            self.__dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(self.__parameter_stdin)
        except pmm_dump_formatter.DumpSpecificationFormatException:
            # parameter_stdin is session-id
            self.__session_id = self.__parameter_stdin
            self.__dump_specification_formatter = None
        if self.__dump_specification_formatter:
            file_path = self.__dump_specification_formatter.get_destination_file()
            try:
                # copy dump to newly created session dir
                session = pmmcli_session.PmmcliSession.createSession('','',file_path,self.__dump_specification_formatter.buildString())
            except IOError, e:
                _logger.info("Error:\n" + str(e))
                return None, 1, 'Unable to initialize session from file ' + file_path + ': ' + e.strerror
        else:
            #get dump from session dir
            session = pmmcli_session.PmmcliSession.openSession(self.__session_id)
        # start pmmcli_daemon
        pmmcli_daemon_service.PMMCliDaemon().start()
        # initialize dump overview maker and get dump-overview
        dump_overview_object = session.getDumpOverview()
        errcode = 0
        data = Data.factory( dump_overview = dump_overview_object)
        return data, errcode, None


class CheckDumpAction(PMMCliAction):
    def __init__(self, parameter_stdin, parameters):
        PMMCliAction.__init__(self, parameter_stdin, parameters)
        self.__parameter_stdin = parameter_stdin
        self.__dump_specification_formatter = None

    def validate(self):
        try:
            error_code, msg = self.get_input_validator().do(self.__parameter_stdin)
        except libxml2.parserError, ex:
            return None, 'XML parse error: ' + ex.msg + '\n' + libxml2errorHandlerErr
        if error_code:
            return None, 'Error ' + str(error_code) + ': ' + msg
        try:
            self.__dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(self.__parameter_stdin)
        except pmm_dump_formatter.DumpSpecificationFormatException:
            return None, 'DumpSpecificationFormatException in DumpSpecification'
        return 1, ''

    def doActivity(self):
        if not self.__dump_specification_formatter:
            self.__dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(self.__parameter_stdin)
        dumps_storage_credentials_formatter = self.__dump_specification_formatter.get_dumps_storage_credentials_formatter()
        dumps_storage_credentials = dumps_storage_credentials_formatter.buildXml()
        storage_type = dumps_storage_credentials_formatter.get_storage_type()
        data = None
        errcode = 0
        message = None
        access_service = None
        if storage_type in ['foreign-ftp','file']:
            access_service = pmm_repository_access_service.FtpRepositoryAccessService
        else:
            access_service = pmm_repository_access_service.LocalRepositoryAccessService
        dump_result, errcode, message = access_service(dumps_storage_credentials_formatter).getDumpStatus(self.__dump_specification_formatter)
        if dump_result:
            dump_status_array = dump_result.get_dump_status()
            check_dump_result_object = CheckDumpResult.factory( dump_status = dump_status_array )
            data = Data.factory( check_dump_result = check_dump_result_object )
        return data, errcode, message


class ExportDumpAsFileAction(PMMCliAction):
    def __init__(self, parameter_stdin, parameters):
        PMMCliAction.__init__(self, parameter_stdin, parameters)
        self.__parameter_stdin = parameter_stdin
        self.__src_dst_files_specification = None

    def validate(self):
        try:
            error_code, msg = self.get_input_validator().do(self.__parameter_stdin)
        except libxml2.parserError, ex:
            return None, 'XML parse error: ' + ex.msg + '\n' + libxml2errorHandlerErr
        if error_code:
            return None, 'Error ' + str(error_code) + ': ' + msg
        try:
            self.__src_dst_files_specification = SrcDstFilesSpecification.factory()
            self.__src_dst_files_specification.build(minidom.parseString(self.__parameter_stdin).childNodes[0])
            if not self.__src_dst_files_specification.get_src() or not self.__src_dst_files_specification.get_dst():
                return None, "SrcDstFilesSpecification has wrong format"
            if self.__src_dst_files_specification.get_src().get_dumps_storage_credentials().get_storage_type() != 'local':
                return None, "ExportDumpAsFileAction called for '" + self.__src_dst_files_specification.get_src().get_dumps_storage_credentials().get_storage_type() + "' storage type"
        except ExpatError:
            return None, 'ExpatError in SrcDstFilesSpecification'
        return 1, ''

    def doActivity(self):
        if not self.__src_dst_files_specification:
            self.__src_dst_files_specification = SrcDstFilesSpecification.factory()
            self.__src_dst_files_specification.build(minidom.parseString(self.__parameter_stdin).childNodes[0])
        src_dumps_specification = self.__src_dst_files_specification.get_src()
        dst_dumps_specification = self.__src_dst_files_specification.get_dst()
        source_dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(src_dumps_specification)
        destination_dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(dst_dumps_specification)
        destination_dumps_storage_credentials_formatter = destination_dump_specification_formatter.get_dumps_storage_credentials_formatter()
        
        storage_type = destination_dumps_storage_credentials_formatter.get_storage_type()
        if storage_type == 'foreign-ftp':
            ftp_password = destination_dumps_storage_credentials_formatter.get_password()
            os.environ['FTP_PASSWORD'] = str(ftp_password)
        
        if mswindows:
            source_dump_storage_credentials = source_dump_specification_formatter.get_dumps_storage_credentials_formatter().buildXml()
            pmm_ras = pmm_repository_access_service.FtpRepositoryAccessService(source_dump_storage_credentials)
            errcode, message, filename = pmm_ras.export_dump_as_file(source_dump_specification_formatter,destination_dump_specification_formatter)
            return None, errcode, message
        
        else:
            cmd = subproc.CmdLine(pmm_config.backup(), self)
            for backup_arg in pmm_config.plesk_backup_args():
                cmd.arg(backup_arg)
            cmd.arg('export-dump-as-file')
            dump_file_name_arg = '--dump-file-name=' + source_dump_specification_formatter.get_destination_file()
            cmd.arg(dump_file_name_arg)
            # output file format is one of these:
            # ftp://login@hostname/root_dir/file_name
            # /root_dir/file_name
            output_file_arg = '--output-file=' + destination_dump_specification_formatter.get_destination_file()
            cmd.arg(output_file_arg)
            
            use_passive_ftp_mode = destination_dumps_storage_credentials_formatter.get_use_passive_ftp_mode()
            if use_passive_ftp_mode == 'true':
                cmd.arg('--ftp-passive-mode')
            
            errcode = 0
            message = u''
            try:
                proc = cmd.spawn()
                #message = u"== STDOUT ====================\n" + proc.stdout + u"\n==============================\n" + u"== STDERR ====================\n" + proc.stderr + u"\n==============================\n"
            except subproc.NonzeroExitException, x:
                errcode = x.exitcode
                message =  u"export-dump-as-file error (Error code = " + unicode(errcode) + "):"  \
                         + u"\n== STDOUT ====================\n" + x.subprocess.stdout + u"\n==============================\n" \
                         + u"== STDERR ====================\n" + x.subprocess.stderr + u"\n==============================\n"
            return None, errcode, message


class ExportFileAsFileAction(PMMCliAction):
    def __init__(self, parameter_stdin, parameters):
        PMMCliAction.__init__(self, parameter_stdin, parameters)
        self.__parameter_stdin = parameter_stdin
        self.__src_dst_files_specification = None

    def validate(self):
        try:
            error_code, msg = self.get_input_validator().do(self.__parameter_stdin)
        except libxml2.parserError, ex:
            return None, 'XML parse error: ' + ex.msg + '\n' + libxml2errorHandlerErr
        if error_code:
            return None, 'Error ' + str(error_code) + ': ' + msg
        try:
            self.__src_dst_files_specification = SrcDstFilesSpecification.factory()
            self.__src_dst_files_specification.build(minidom.parseString(self.__parameter_stdin).childNodes[0])
            if not self.__src_dst_files_specification.get_src() or not self.__src_dst_files_specification.get_dst():
                return None, "SrcDstFilesSpecification has wrong format"
            if self.__src_dst_files_specification.get_src().get_dumps_storage_credentials().get_storage_type() == 'local':
                return None, "ExportFileAsFileAction called for 'local' source storage type"
            if self.__src_dst_files_specification.get_dst().get_dumps_storage_credentials().get_storage_type() == 'local':
                return None, "ExportFileAsFileAction called for 'local' destination storage type"
        except ExpatError:
            return None, 'ExpatError in SrcDstFilesSpecification'
        return 1, ''

    def doActivity(self):
        if not self.__src_dst_files_specification:
            self.__src_dst_files_specification = SrcDstFilesSpecification.factory()
            self.__src_dst_files_specification.build(minidom.parseString(self.__parameter_stdin).childNodes[0])
        src_dumps_specification = self.__src_dst_files_specification.get_src()
        dst_dumps_specification = self.__src_dst_files_specification.get_dst()
        source_dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(src_dumps_specification)
        destination_dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(dst_dumps_specification)
        source_dump_storage_credentials = source_dump_specification_formatter.get_dumps_storage_credentials_formatter().buildXml()
        pmm_ras = pmm_repository_access_service.FtpRepositoryAccessService(source_dump_storage_credentials)
        errcode, message = pmm_ras.export_file_as_file(destination_dump_specification_formatter)
        return None, errcode, message


class ImportFileAsDumpAction(PMMCliAction):
    def __init__(self, parameter_stdin, parameters):
        PMMCliAction.__init__(self, parameter_stdin, parameters)
        self.__parameter_stdin = parameter_stdin
        self.__src_dst_files_specification = None

    def validate(self):
        try:
            error_code, msg = self.get_input_validator().do(self.__parameter_stdin)
        except libxml2.parserError, ex:
            return None, 'XML parse error: ' + ex.msg + '\n' + libxml2errorHandlerErr
        if error_code:
            return None, 'Error ' + str(error_code) + ': ' + msg
        try:
            self.__src_dst_files_specification = SrcDstFilesSpecification.factory()
            self.__src_dst_files_specification.build(minidom.parseString(self.__parameter_stdin).childNodes[0])
            if not self.__src_dst_files_specification.get_src() or not self.__src_dst_files_specification.get_dst():
                return None, "SrcDstFilesSpecification has wrong format"
            if self.__src_dst_files_specification.get_dst().get_dumps_storage_credentials().get_storage_type() != 'local':
                return None, "ExportDumpAsFileAction called for '" + self.__src_dst_files_specification.get_dst().get_dumps_storage_credentials().get_storage_type() + "' storage type"
            owner_guid = self.__src_dst_files_specification.get_guid()
            owner_type = self.__src_dst_files_specification.get_type()
            if owner_type == None:
                return None, "SrcDstFilesSpecification has wrong format: 'type' attribute is required for '--import-file-as-dump' command'"
            if owner_guid == None:
                return None, "SrcDstFilesSpecification has wrong format: 'guid' attribute is required for '--import-file-as-dump' command'"
            if owner_guid == '':
                return None, "SrcDstFilesSpecification has wrong format: not-empty 'guid' attribute is required for '--import-file-as-dump' command'"
        except ExpatError:
            return None, 'ExpatError in SrcDstFilesSpecification'
        return 1, ''

    def doActivity(self):
        if not self.__src_dst_files_specification:
            self.__src_dst_files_specification = SrcDstFilesSpecification.factory()
            self.__src_dst_files_specification.build(minidom.parseString(self.__parameter_stdin).childNodes[0])
        src_dumps_specification = self.__src_dst_files_specification.get_src()
        dst_dumps_specification = self.__src_dst_files_specification.get_dst()
        owner_guid = self.__src_dst_files_specification.get_guid()
        owner_type = self.__src_dst_files_specification.get_type()
        source_dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(src_dumps_specification)
        destination_dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(dst_dumps_specification)
        source_dump_storage_credentials = source_dump_specification_formatter.get_dumps_storage_credentials_formatter().buildXml()
        pmm_ras = pmm_repository_access_service.FtpRepositoryAccessService(source_dump_storage_credentials)
        errcode_1, message, filename = pmm_ras.import_file_as_dump(source_dump_specification_formatter,destination_dump_specification_formatter,owner_guid,owner_type)
        _logger.debug("Import file operation ended with errcode=%d. Filename is %s" % (errcode_1, filename))
        # error code 111 returned if such dump already exists in destination repository
        # error code 112 returned if object specified by guid and type has lower level than backup type
        
        if errcode_1 != 0:
            if errcode_1 == 111:
                message = u"Dump already exists in repository"
            else:
                if errcode_1 == 112:
                    message = unicode("The backup you are trying to import does not match the user type '%s'" % owner_type)
                return None, errcode_1, message
        
        result_message = message
        # check dump
        destination_dump_specification_formatter.set_name_of_xml_file(filename)
        destination_dumps_storage_credentials_formatter = destination_dump_specification_formatter.get_dumps_storage_credentials_formatter()
        dump_result, errcode_2, message = pmm_repository_access_service.LocalRepositoryAccessService(destination_dumps_storage_credentials_formatter).getDumpStatus(destination_dump_specification_formatter)
        _logger.debug("Check dump operation ended.")
        if message:
            result_message = result_message + u"pmm-ras check dump error:\n" + unicode(message)
        
        data = Data.factory( dump = dump_result )
        
        errcode = None
        if errcode_2 == 0 and errcode_1 == 111:
            errcode = errcode_1
        else:
            errcode = errcode_2
        
        return data, errcode, result_message


class RestoreAction(PMMCliAction):
    def __init__(self, parameter_stdin, parameters):
        PMMCliAction.__init__(self, parameter_stdin, parameters)
        self.__parameter_stdin = parameter_stdin
        self.__restore_task_description = None

    def validate(self):
        try:
            error_code, msg = self.get_input_validator().do(self.__parameter_stdin)
        except libxml2.parserError, ex:
            return None, 'XML parse error: ' + ex.msg + '\n' + libxml2errorHandlerErr
        if error_code:
            return None, 'Error ' + str(error_code) + ': ' + msg
        try:
            self.__restore_task_description = RestoreTaskDescription.factory()
            self.__restore_task_description.build(minidom.parseString(self.__parameter_stdin).childNodes[0])
        except ExpatError:
            return None, 'ExpatError in RestoreTaskDescription'
        # require both of owner_guid and owner_type specified for file restore
        owner_guid = self.__restore_task_description.get_owner_guid()
        owner_type = self.__restore_task_description.get_owner_type()
        if not owner_guid or not owner_type or owner_guid == '' or owner_type == '':
             return None, "Could not restore dump: owner-guid and owner-type are required"
        return 1, ''

    def addDaemonTaskDeleteDump(self, dump_specification_formatter):
        task_dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(dump_specification_formatter)
        daemon_task = DaemonTaskDeleteDump(task_dump_specification_formatter)
        try:
            daemon_task.save()
        except Exception, e:
            _logger.error("Exception %s occured during save daemon task to delete dump: %s\n Call stack is:\n%s" %(str(e.__class__),str(e),stacktrace.stacktrace()))

    def doActivity(self):
        if not self.__restore_task_description:
            self.__restore_task_description = RestoreTaskDescription.factory()
            self.__restore_task_description.build(minidom.parseString(self.__parameter_stdin).childNodes[0])
        ignore_sign = None
        ignore_conflicts = None
        ignore_errors = None
        if self.__restore_task_description.get_ignore_errors():
            ignore_errors = self.__restore_task_description.get_ignore_errors().get_ignore_error()
            for ignore_error in ignore_errors:
                if ignore_error.get_type() == 'conflicts':
                    ignore_conflicts = 1
                if ignore_error.get_type() == 'sign':
                    ignore_sign = 1
        
        selected_objects = self.__restore_task_description.get_objects().get_selected()
        
        conflict_resolution_rules = self.__restore_task_description.get_conflict_resolution_rules()
        
        add_daemon_delete_dump_task = False
        
        owner_guid = self.__restore_task_description.get_owner_guid()
        owner_type = self.__restore_task_description.get_owner_type()
        
        dump_specification_formatter = None
        imported_dump_specification_formatter = None
        d_s = self.__restore_task_description.get_source().get_dump_specification()
        if d_s:
            # source is dump specification so we have to create new session
            dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(d_s)
            storage_type = dump_specification_formatter.get_dumps_storage_credentials_formatter().get_storage_type()
            if storage_type == 'local': 
                session = pmmcli_session.PmmcliSession.createSession(owner_type,owner_guid,dump_specification_formatter.get_destination_file(),dump_specification_formatter.buildString())
            else:
                # now pmmcli should make import-file-as-dump operation, check dump and then restore it if dump is OK
                # should initialize destination_dump_specification_formatter with some name to avoid exception 'empty name with local storage-type'
                destination_dump_specification_formatter = pmm_dump_formatter.getLocalDumpSpecificationFormatter('dummy.xml')
                source_dump_storage_credentials = dump_specification_formatter.get_dumps_storage_credentials_formatter().buildXml()
                pmm_ras = pmm_repository_access_service.FtpRepositoryAccessService(source_dump_storage_credentials)
                errcode, message, imported_dump_filename = pmm_ras.import_file_as_dump(dump_specification_formatter,destination_dump_specification_formatter,owner_guid,owner_type)
                _logger.debug("Import file operation ended with errcode=%d. Filename is %s" % (errcode, imported_dump_filename))
                # error code 111 returned if such dump already exists in destination repository
                # error code 112 returned if object specified by guid and type have no rights to import dump
                # anyway we should continue only if errcode is 0 or 111
                if errcode != 0:
                    if errcode == 111:
                        message = u"Dump already exists in repository"
                    else:
                        if errcode == 112:
                            message = unicode("The backup you are trying to import does not match the user type '%s'" % owner_type)
                        return None, errcode, message
                if errcode == 0:
                    add_daemon_delete_dump_task = True
                
                imported_dump_specification_formatter = pmm_dump_formatter.getLocalDumpSpecificationFormatter(imported_dump_filename.lstrip('/'))
                # call check dump for imported dump
                imported_dump_specification_formatter.set_name_of_xml_file(imported_dump_filename)
                imported_dump_storage_credentials_formatter = imported_dump_specification_formatter.get_dumps_storage_credentials_formatter()
                dump_result, errcode, message = pmm_repository_access_service.LocalRepositoryAccessService(imported_dump_storage_credentials_formatter).getDumpStatus(imported_dump_specification_formatter)
                _logger.debug("Check dump operation ended.")
                # errcode always is None
                # we should analyse dump_result to get dump status
                status_array = dump_result.get_dump_status()
                dump_failed_on_status = False
                for status in status_array:
                    _logger.debug("Import dump status: %s" % status.get_dump_status())
                    if status.get_dump_status() in ['WRONG-FORMAT', 'CONTENT-ERROR']:
                        dump_failed_on_status = True
                
                result_message = u''
                if dump_failed_on_status:
                    if message:
                        result_message = u"pmm-ras check dump error:\n" + unicode(message)
                    if add_daemon_delete_dump_task and imported_dump_specification_formatter != None:
                        self.addDaemonTaskDeleteDump(imported_dump_specification_formatter)
                    return None, 14, u"Check imported dump " + unicode(imported_dump_filename) + u" failed.\n" + result_message
                
                # dump is imported and checked
                session = pmmcli_session.PmmcliSession.createSession(owner_type,owner_guid,imported_dump_specification_formatter.get_destination_file(),imported_dump_specification_formatter.buildString())
                
        else:
            # create new session based on selected objects
            session_id = self.__restore_task_description.get_source().get_session_id()
            source_session = pmmcli_session.PmmcliSession(session_id)
            if source_session.conflicts_resolved():
                session = pmmcli_session.PmmcliSession.cloneSession(session_id)
            else:
                session = pmmcli_session.PmmcliSession.deriveSession(session_id)
            dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(session.get_dump_specification_value())
        
        # session is created
        validity_error = None
        result = RestoreTaskResult()
        
        # omit sign checker if conflicts already are resolved
        if not session.conflicts_resolved():
            _logger.debug("Prepare to call sign checker")
            # invoke sign checker, validate result and generate error if necessary
            backup_sign_errcode = session.checkSign()
            if not ignore_sign:
                if backup_sign_errcode != 0:
                    backup_sign_error_codes = {1: "no integrity", 2: "bad signature",3: "not signed"}
                    validity_error = DumpValidityError.factory()
                    validity_error.set_type(str(backup_sign_errcode))
                    backup_sign_error_message = backup_sign_error_codes.get(backup_sign_errcode,'unexpected error from backup-sign') 
                    validity_error.setValueOf_(backup_sign_error_message)
                    result.set_dump_validity_error([validity_error])
                    fatal_error_codes = [1,2]
                    if backup_sign_errcode in fatal_error_codes:
                        errcode = 105
                        data = Data.factory( restore_task_result = result )
                        if add_daemon_delete_dump_task and imported_dump_specification_formatter != None:
                            self.addDaemonTaskDeleteDump(imported_dump_specification_formatter)
                        return data, errcode, "Bad dump: Check sign error: " + backup_sign_error_message
            # now 'restore.xml' is unsigned and it is the time adjust 'base'
            session.apply_base()
        
        session.create_restore_specification(selected_objects,owner_type,owner_guid)
        
        # at this time 'restore.xml' file is created and ready to deploy
        if not ignore_conflicts:
            _logger.debug("Prepare to detect conflicts")
            # work with conflict detector
            plesk_actual_dump = PleskActualDump(session.get_session_path()).get()
            if session.detectConflicts(plesk_actual_dump):
                if conflict_resolution_rules:
                    packet = convertToXmlString(conflict_resolution_rules,'conflict-resolution-rules')
                    conflicts_found = session.resolveConflicts(plesk_actual_dump,packet)
                if not conflict_resolution_rules or conflicts_found != 0:
                    errcode = 0
                    overview = session.getDumpOverview()
                    conflicts = session.getConflictDescription()
                    result.set_dump_overview(overview)
                    result.set_conflicts_description(conflicts)
                    data = Data.factory( restore_task_result = result )
                    if add_daemon_delete_dump_task and imported_dump_specification_formatter != None:
                        self.addDaemonTaskDeleteDump(imported_dump_specification_formatter)
                    return data, errcode, None
        
        # start deployer
        _logger.debug("Prepare to start deployer")
        cmd = subproc.CmdLine(pmm_config.deployer(), stderr = session.create_log('.stderr'))
        cmd.arg('--deploy-dump')
        if self.__restore_task_description.get_misc().get_suspend() == "true":
            cmd.arg('--suspend')
        
        verbose_level = self.__restore_task_description.get_misc().get_verbose_level()
        try:
            verbose_level_value = int(verbose_level)
            if verbose_level_value >= 1:
                cmd.arg('--verbose')
        except (TypeError, ValueError):
            pass
        
        cmd.arg('--session-path')
        session_path_arg = session.get_session_path()
        cmd.arg(session_path_arg)
        
        # Get restore tasks list
        #
        # Need additional protection against two pmmcli instance run
        # because walking throw Task List takes some time.
        # Use osutil.Interlock instance to lock access to code section
        _logger.debug("Acquire deployer lock")
        mutex = osutil.Interlock(pmm_config.pmmcli_deployer_lock_file())
        if not mutex.lock():
            _logger.debug('Deployer lock already locked')
            return None, 201, "Unable to start Deployer because there is another Deployer run"
        
        _logger.debug("Deployer lock acquired")
        task_manager = pmm_task.getPMMTaskManager()
        restore_task_list = task_manager.realGetTaskList('Restore')
        for task in restore_task_list:
            task_status = task_manager.operatorGetTaskStatus(task)
            if task_status.get_working():
                _logger.debug("Working restore task detected (%s)" % str(task.get_pmm_pid()))
                if add_daemon_delete_dump_task and imported_dump_specification_formatter != None:
                    self.addDaemonTaskDeleteDump(imported_dump_specification_formatter)
                mutex.unlock()
                return None, 201, "Unable to start Deployer because there is another Deployer run"
        
        errcode = 0
        try:
            mutex.unlock()
            _logger.debug("Deployer lock unlocked")
            _logger.debug("Start deployer")
            subprocess = cmd.asyncSpawn()
            pid = subprocess.get_pid()
        except subproc.NonzeroExitException, x:
            errcode = x.exitcode
        
        storage_full_name = dump_specification_formatter.get_dumps_storage_credentials_formatter().getDumpStorage()
        name = session.get_info_xml_base().replace(storage_full_name,'')
        full_name = name
        
        task = pmm_task.RestoreTask( cmd.get_cmd(), pid, self.__restore_task_description.get_owner_guid(),self.__restore_task_description.get_owner_type(), session.get_session_path(),dump_specification_formatter.get_dumps_storage_credentials_formatter().buildString(),full_name,name,self.__restore_task_description.get_mail_to())
        task_id_object = TaskId()
        task_id_object.setValueOf_(task.get_pmm_pid())
        data = Data.factory( task_id = task_id_object )
        if add_daemon_delete_dump_task and imported_dump_specification_formatter != None:
            self.addDaemonTaskDeleteDump(imported_dump_specification_formatter)
        return data, errcode, None


class MakeDumpAction(PMMCliAction):
    def __init__(self, parameter_stdin, parameters):
        PMMCliAction.__init__(self, parameter_stdin, parameters)
        self.__parameter_stdin = parameter_stdin
        self.__backup_task_description = None
        self.__dumps_storage_credentials_formatter = None

    def generateSessionPath(self):
        pmm_session_path = pmm_config.session_dir()
        if not os.path.isdir(pmm_session_path):
            os.mkdir(pmm_session_path)
        session_path = os.path.join(pmm_config.session_dir(),time.strftime('%Y-%m-%d-%H%M%S.',time.localtime()) + str(random.randint(0,1000)))
        while os.path.isdir(session_path):
            session_path = os.path.join(pmm_config.session_dir(),time.strftime('%Y-%m-%d-%H%M%S.',time.localtime())+str(random.randint(0,1000)))
        os.mkdir(session_path)
        return session_path

    def validate(self):
        try:
            error_code, msg = self.get_input_validator().do(self.__parameter_stdin)
        except libxml2.parserError, ex:
            return None, 'XML parse error: ' + ex.msg + '\n' + libxml2errorHandlerErr
        if error_code:
            return None, 'Error ' + str(error_code) + ': ' + msg
        try:
            self.__backup_task_description = BackupTaskDescription.factory()
            self.__backup_task_description.build(minidom.parseString(self.__parameter_stdin).childNodes[0])
        except ExpatError:
            return None, 'ExpatError in BackupTaskDescription'
        try:
            self.__backup_task_description.get_misc().get_backup_profile_name().encode('ascii')
        except UnicodeEncodeError:
            return None, 'backup-profile-name must be ascii string'
        self.__dumps_storage_credentials_formatter = pmm_dump_formatter.DumpsStorageCredentialsFormatter(self.__backup_task_description.get_dumps_storage_credentials())
        commands_to_backup = self.convertBackupSpecification2Args(self.__backup_task_description.get_backup_specification())[0]
        if len(commands_to_backup) == 0:
            return None, 'Null-length command-to-backup in ExpatError in BackupSpecification'
        return 1, ''

    def convertBackupSpecification2Args(self, backup_specification):
        object_to_backup_list = backup_specification.get_object_to_backup()
        object_to_exclude_list = backup_specification.get_object_to_exclude()
        
        # list of allowed types&backup_commands sorted by prority
        allowed_types = ['server','reseller','client','domain']
        # list of the types, enumerable by ids or names 
        enum_types = ['reseller','client','domain']
        
        commands_to_backup = []
        objects_to_backup_args = []
        objects_to_backup_warnings = []
        
        backup_command_index = None
        
        backup_all = False
        
        # Lists are structured as followings: [enum_types('reseller','client','domain'):fixed length] X [(name,id):fixed length] X [objects list:growing length]
        # 'backup_objects[0][1][0]' is an id of the first reseller (given by its id) and
        # 'backup_objects[1][0][2]' is a name of the third client (given by its name) if BackupSpecification has clent names and ids mixed (specification v.0.20 allow this)
        backup_objects = [[[] for col in range(2)] for row in range(len(enum_types))]
        exclude_objects = [[[] for col in range(2)] for row in range(len(enum_types))]
        
        backup_objects_all = [0 for row in range (len(enum_types))]
        exclude_objects_all = [0 for row in range (len(enum_types))]
        
        for object_item in object_to_backup_list:
            object_type = object_item.get_type()
            if object_type in allowed_types:
                if (backup_command_index ==None) or (allowed_types.index(object_type) < backup_command_index):
                    backup_command_index = allowed_types.index(object_type)
                if allowed_types.index(object_type) != 0:
                    if object_item.get_all() == 'true':
                        backup_objects_all[enum_types.index(object_type)] = 1
                    else:
                        object_id = object_item.get_id()
                        if object_id == '':
                            object_id = None
                        object_name = object_item.get_name()
                        if object_name == '':
                            object_name = None
                        if object_id:
                            backup_objects[enum_types.index(object_type)][1].append(object_id)
                        elif object_name:
                            backup_objects[enum_types.index(object_type)][0].append(object_name)
                        else:
                            objects_to_backup_warnings.append("Backup Task warning: Object to backup specified by type='%s' has neither 'id' nor 'name' specified.\n" % (object_type))
        
        for object_item in object_to_exclude_list:
            object_type = object_item.get_type()
            if object_type in allowed_types:
                if allowed_types.index(object_type) == 0:
                    objects_to_backup_warnings.append("Backup Task warning: 'Exclude %s' is not supported option\n" % (allowed_types[0]))
                else:
                    if object_item.get_all() == 'true':
                        exclude_objects_all[enum_types.index(object_type)] = 1
                    else:
                        object_id = object_item.get_id()
                        if object_id == '':
                            object_id = None
                        object_name = object_item.get_name()
                        if object_name == '':
                            object_name = None
                        if object_id:
                            exclude_objects[enum_types.index(object_type)][1].append(object_id)
                        elif object_name:
                            exclude_objects[enum_types.index(object_type)][0].append(object_name)
                        else:
                            objects_to_backup_warnings.append("Backup Task warning: Object to exclude specified by type='%s' has neither 'id' nor 'name' specified.\n" % (object_type))
        if backup_command_index == None:
            objects_to_backup_warnings.append("Backup Task warning: Could not find backup command. '%s' command will be used instead" % (allowed_types[0]))
            commands_to_backup.append(allowed_types[0])
        elif backup_command_index == 0:
            commands_to_backup.append(allowed_types[0])
        else:
            if backup_objects_all[backup_command_index-1]:
                commands_to_backup.append(allowed_types[backup_command_index]+'s-id')
                backup_all = True 
                # no id agrs should be specifies in this case, this forces backing up all the clients/domains/etc
            else:
                if len(backup_objects[backup_command_index-1][1]) > 0:
                    commands_to_backup.append(allowed_types[backup_command_index]+'s-id')
                    for item_id in backup_objects[backup_command_index-1][1]:
                        objects_to_backup_args.append(item_id)
                    for item_name in backup_objects[backup_command_index-1][0]:
                        objects_to_backup_warnings.append("Backup Task warning: %s '%s' does not included into backup. Cannot specify %s's ids and names together.\n" % (allowed_types[backup_command_index], item_name, allowed_types[backup_command_index]))
                
                else:
                    commands_to_backup.append(allowed_types[backup_command_index]+'s-name')
                    for item_name in backup_objects[backup_command_index-1][0]:
                        objects_to_backup_args.append(item_name)
                
            rest_alls = backup_objects_all[backup_command_index:]
            for i in range(len(rest_alls)):
                if rest_alls[i]:
                    objects_to_backup_warnings.append("Backup Task warning: 'All %ss' does not included into backup being shadowed with backing up '%ss'.\n" % (enum_types[backup_command_index+i], allowed_types[backup_command_index]))
            
            for i in range(len(backup_objects) - backup_command_index):
                for item_name in backup_objects[i + backup_command_index][0]:
                    objects_to_backup_warnings.append("Backup Task warning: %s '%s' does not included into backup being shadowed with backing up '%ss'.\n" % (allowed_types[i+ backup_command_index], item_name, allowed_types[backup_command_index]))
                for item_id in backup_objects[i + backup_command_index][1]:
                    objects_to_backup_warnings.append("Backup Task warning: %s with id=%s does not included into backup being shadowed with backing up '%ss'.\n" % (allowed_types[i+ backup_command_index], item_id, allowed_types[backup_command_index]))
            
            # processing excludes, should forms one command '--exclude-XXXX' by one object type
            for i in range(len(exclude_objects_all)):
                if exclude_objects_all[i]:
                    objects_to_backup_warnings.append("Backup Task warning: 'Exclude all %ss' does not supported." % (enum_types[i]))
            
            for i in range(len(exclude_objects)):
                for item_id in exclude_objects[i][1]:
                    objects_to_backup_warnings.append("Backup Task warning: Exclude %s with id='%s' does not included into backup. Cannot exclude %s by its id.\n" % (enum_types[i], item_id, enum_types[i]))
            
        for i in range(len(exclude_objects)):
            if len(exclude_objects[i][0]) > 0:
                cmd = '--exclude-' + enum_types[i] + '='
                delimiter = ''
                for item_name in exclude_objects[i][0]:
                    cmd = cmd + delimiter + item_name
                    delimiter = ','
                commands_to_backup.append(cmd)
        
        warning_message = ''.join(objects_to_backup_warnings)
        return commands_to_backup, backup_all, objects_to_backup_args, warning_message

    def doActivity(self):
        if not self.__backup_task_description:
            self.__backup_task_description = BackupTaskDescription.factory()
            self.__backup_task_description.build(minidom.parseString(self.__parameter_stdin).childNodes[0])
        if not self.__dumps_storage_credentials_formatter:
            self.__dumps_storage_credentials_formatter = pmm_dump_formatter.DumpsStorageCredentialsFormatter(self.__backup_task_description.get_dumps_storage_credentials())
        
        session_path = self.generateSessionPath()
        
        # topobject_to_backup always exists, but may have invalid format
        topobject_to_backup = self.__backup_task_description.get_backup_specification().get_object_to_backup()[0]
        topobject_type = topobject_to_backup.get_type()
        topobject_id = topobject_to_backup.get_id()
        
        # check available disk space
        disk_space_required = pmmcli_config.get().get_free_disk_space()
        local_dump_directory = plesk_config.get("DUMP_D")
        if local_dump_directory:
            if osutil.free_bytes(local_dump_directory) < (long(disk_space_required) * 1048576):
                task = pmm_task.BackupTask( '', -1, self.__backup_task_description.get_misc().get_owner_guid(),self.__backup_task_description.get_misc().get_owner_type(), session_path, self.__dumps_storage_credentials_formatter.buildString(),self.__backup_task_description.get_misc().get_backup_profile_name(),None,topobject_id,topobject_type)
                task_id_object = TaskId()
                task_id_object.setValueOf_(task.get_pmm_pid())
                data = Data.factory( task_id = task_id_object )
                error_message = "Low Disk space for backup"
                return data, 11, error_message
        
        ftp_output_file = None
        output_file = None
        use_passive_ftp_mode = None
        
        storage_type = self.__dumps_storage_credentials_formatter.get_storage_type()
        if storage_type in ['foreign-ftp','file']:
            output_file = self.__dumps_storage_credentials_formatter.getFileDumpStorage()
            if storage_type == 'foreign-ftp':
                ftp_password = self.__dumps_storage_credentials_formatter.get_password()
                os.environ['FTP_PASSWORD'] = str(ftp_password)
                if self.__dumps_storage_credentials_formatter.get_use_passive_ftp_mode() == 'true':
                    use_passive_ftp_mode = True
        
        commands_to_backup, backup_all, objects_to_backup_args, objects_to_backup_warning_message = self.convertBackupSpecification2Args(self.__backup_task_description.get_backup_specification())
        
        if len(commands_to_backup) >= 0 and commands_to_backup[0] != "server" and not backup_all and len(objects_to_backup_args) == 0:
            return None, 12, "Could not backup specified object: wrong object-to-backup format:\n" + objects_to_backup_warning_message
        
        cmd = subproc.CmdLine(pmm_config.backup(), stderr = open(os.path.join(session_path,'.stderr'),'wt'))
        for backup_arg in pmm_config.plesk_backup_args():
            cmd.arg(backup_arg)
        
        for backup_cmd in commands_to_backup:
            cmd.arg(backup_cmd)
        
        owner_guid_arg = '--owner-uid=' + self.__backup_task_description.get_misc().get_owner_guid()
        cmd.arg(owner_guid_arg)
        owner_type = self.__backup_task_description.get_misc().get_owner_type()
        if owner_type:
            owner_type_arg = '--owner-type=' + owner_type
            cmd.arg(owner_type_arg)
        
        dump_rotation = self.__backup_task_description.get_backup_specification().get_backup_options().get_rotation()
        if dump_rotation:
            dump_rotation_arg = '--dump-rotation=' + dump_rotation
            cmd.arg(dump_rotation_arg)
        
        backup_profile_id = self.__backup_task_description.get_misc().get_backup_profile_id()
        if backup_profile_id != '':
            backup_profile_id_arg = '--backup-profile-id=' + backup_profile_id
            cmd.arg(backup_profile_id_arg)
        
        # TODO: wrap backup profile name into "" when pleskbackup will understand this
        backup_profile_name = self.__backup_task_description.get_misc().get_backup_profile_name()
        if backup_profile_name != '':
            backup_profile_name_arg = '--backup-profile-name=' + backup_profile_name
            cmd.arg(backup_profile_name_arg)
        
        split_size = self.__backup_task_description.get_backup_specification().get_backup_options().get_split_size()
        if split_size:
            split_arg = '--split=' + split_size
            cmd.arg(split_arg)
        
        dump_description = self.__backup_task_description.get_backup_specification().get_backup_options().get_description()
        if dump_description:
            dump_description_arg = '--description=' + dump_description
            cmd.arg(dump_description_arg)
        
        if self.__backup_task_description.get_backup_specification().get_backup_options().get_type() == "configuration-only":
            cmd.arg('-c')
        
        if self.__backup_task_description.get_backup_specification().get_backup_options().get_compression_level() == "do-not-compress":
            cmd.arg('--no-gzip')
        
        backup_filter = self.__backup_task_description.get_backup_specification().get_backup_options().get_filter()
        if backup_filter == "only-mail":
            cmd.arg('--only-mail')
        elif backup_filter == "only-phosting":
            cmd.arg('--only-hosting')
            
        if self.__backup_task_description.get_backup_specification().get_backup_options().get_suspend() == "true":
            cmd.arg('--suspend')
        
        verbose_level = self.__backup_task_description.get_misc().get_verbose_level()
        try:
            verbose_level_value = int(verbose_level)
            if verbose_level_value >= 1:
                if verbose_level_value > 4:
                    verbose_level_value = 4
                verbose_arg = '-' + 'v'*verbose_level_value
                cmd.arg(verbose_arg)
        except (TypeError, ValueError):
            pass
        
        # TODO: wrap session path into "" when pleskbackup will understand this
        session_path_arg = '--session-path=' + session_path
        cmd.arg(session_path_arg)
        
        if output_file:
            output_file_arg = '--output-file=' + output_file
            cmd.arg(output_file_arg)
        
        if use_passive_ftp_mode:
            cmd.arg('--ftp-passive-mode')
        
        # enumerate object names or ids 
        for object_to_backup_arg in objects_to_backup_args:
            cmd.arg(object_to_backup_arg)
        
        
        errcode = 0
        data = None
        try:
            subprocess = cmd.asyncSpawn()
            pid = subprocess.get_pid()
            task = pmm_task.BackupTask( cmd.get_cmd(), pid, self.__backup_task_description.get_misc().get_owner_guid(),self.__backup_task_description.get_misc().get_owner_type(), session_path, self.__dumps_storage_credentials_formatter.buildString(),self.__backup_task_description.get_misc().get_backup_profile_name(),self.__backup_task_description.get_backup_specification().get_backup_options().get_mail_to(),topobject_id,topobject_type)
            task_id_object = TaskId()
            task_id_object.setValueOf_(task.get_pmm_pid())
            data = Data.factory( task_id = task_id_object )
        except subproc.NonzeroExitException, x:
            errcode = x.exitcode
        return data, errcode, objects_to_backup_warning_message


class DeleteDumpAction(PMMCliAction):
    def __init__(self, parameter_stdin, parameters):
        PMMCliAction.__init__(self, parameter_stdin, parameters)
        self.__parameter_stdin = parameter_stdin
        self.__dump_specification_formatter = None

    def validate(self):
        try:
            error_code, msg = self.get_input_validator().do(self.__parameter_stdin)
        except libxml2.parserError, ex:
            return None, 'XML parse error: ' + ex.msg + '\n' + libxml2errorHandlerErr
        if error_code:
            return None, 'Error ' + str(error_code) + ': ' + msg
        try:
            self.__dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(self.__parameter_stdin)
        except pmm_dump_formatter.DumpSpecificationFormatException:
            return None, 'DumpSpecificationFormatException in DumpSpecification'
        return 1, ''

    def doActivity(self):
        if not self.__dump_specification_formatter:
            self.__dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(self.__parameter_stdin)
        dumps_storage_credentials_formatter = self.__dump_specification_formatter.get_dumps_storage_credentials_formatter()
        dumps_storage_credentials = dumps_storage_credentials_formatter.buildXml()
        storage_type = dumps_storage_credentials_formatter.get_storage_type()
        errcode = 0
        message = None
        access_service = None
        if storage_type != 'local':
            access_service = pmm_repository_access_service.FtpRepositoryAccessService
        else:
            access_service = pmm_repository_access_service.LocalRepositoryAccessService
        errcode, message = access_service(dumps_storage_credentials).deleteDump(self.__dump_specification_formatter.buildXml())
        return None, errcode, message


class ResolveConflictsAction(PMMCliAction):
    def __init__(self, parameter_stdin, parameters):
        PMMCliAction.__init__(self, parameter_stdin, parameters)
        self.__parameter_stdin = parameter_stdin
        self.__resolve_conflicts_task_description = None
        
    def get_resolve_conflicts_task_description(self):
        return self.__resolve_conflicts_task_description

    def validate(self):
        try:
            error_code, msg = self.get_input_validator().do(self.__parameter_stdin)
        except libxml2.parserError, ex:
            return None, 'XML parse error: ' + ex.msg + '\n' + libxml2errorHandlerErr
        if error_code:
            return None, 'Error ' + str(error_code) + ': ' + msg
        try:
            self.__resolve_conflicts_task_description = ResolveConflictsTaskDescription.factory()
            self.__resolve_conflicts_task_description.build(minidom.parseString(self.__parameter_stdin).childNodes[0])
        except ExpatError:
            return None, 'ExpatError in ResolveConflictsTaskDescription'
        return 1, ''

    def doActivity(self):
        if not self.__resolve_conflicts_task_description:
            self.__resolve_conflicts_task_description = ResolveConflictsTaskDescription.factory()
            self.__resolve_conflicts_task_description.build(minidom.parseString(self.__parameter_stdin).childNodes[0])
        session_id = self.get_resolve_conflicts_task_description().get_session_id()
        rsession = pmmcli_session.PmmcliSession.cloneSession(session_id)
        packet = convertToXmlString(self.get_resolve_conflicts_task_description().get_conflict_resolution_rules(),'conflict-resolution-rules')
        rsession.resolveConflictsOnce(rsession.get_plesk_actual_dump(),packet)
        dump_overview_object = rsession.getDumpOverview()
        data = Data.factory(dump_overview = dump_overview_object)
        return data, 0, None

class GetConflictsDescriptionAction(PMMCliAction):
    def __init__(self, parameter_stdin, parameters):
        PMMCliAction.__init__(self, parameter_stdin, parameters)
        self.__parameters = parameters
        self.__session_id = None
        
    def validate(self):
        if len(self.__parameters) != 1:
            return None, "Parameter 'session_id' not specified"
        self.__session_id = self.__parameters[0]
        return 1, ''

    def doActivity(self):
        if not self.__session_id:
            self.__session_id = self.__parameters[0]
        errcode = 0
        session = pmmcli_session.PmmcliSession(self.__session_id)
        overview = session.getDumpOverview()
        conflicts = session.getConflictDescription()
        result = RestoreTaskResult(dump_overview = overview, conflicts_description = conflicts)
        data = Data.factory( restore_task_result = result )
        return data, errcode, None

class GetTasksListAction(PMMCliAction):
    def __init__(self, parameter_stdin, parameters):
        PMMCliAction.__init__(self, parameter_stdin, parameters)
        self.__parameters = parameters
        self.__type = ''
        self.__owner_guid = None
        self.__topobject_id = None
        self.__topobject_type = None

    def validate(self):
        if len(self.__parameters) < 1:
            return None, "Parameter 'task_type' not specified"
        if not self.__parameters[0] in ['Backup','Restore']:
            return None, "Task type must be 'Backup' or 'Restore'"
        self.__type = self.__parameters[0]
        if len(self.__parameters) >= 2:
            self.__owner_guid = self.__parameters[1]
        if self.__type == 'Backup':
            if len(self.__parameters) >= 3:
                self.__topobject_id = self.__parameters[2]
            if len(self.__parameters) >= 4:
                if not self.__parameters[3] in ['server','reseller','client','domain']:
                    return None, "Top object type must be one of: 'server','reseller','client','domain'"
                self.__topobject_type = self.__parameters[3]
        return 1, ''

    def doActivity(self):
        data = Data.factory(task_list = pmm_task.getPMMTaskManager().getTaskList(self.__type,self.__owner_guid,self.__topobject_id,self.__topobject_type))
        return data, 0, None


class GetTaskAction(PMMCliAction):
    def __init__(self, parameter_stdin, parameters):
        PMMCliAction.__init__(self, parameter_stdin, parameters)
        self.__parameters = parameters
        self.__task_id = None

    def validate(self):
        if len(self.__parameters) != 1:
            return None, "Parameter 'task_id' not specified"
        self.__task_id = self.__parameters[0]
        # there is no format requirements for task-id
        return 1, ''

    def doActivity(self):
        if not self.__task_id:
            self.__task_id = self.__parameters[0]
        pmm_task_list = pmm_task.getPMMTaskManager().getTask(self.__task_id)
        if pmm_task_list:
            data = Data.factory(task_list = pmm_task_list)
            return data, 0, None
        else:
            return None, 2, 'No task found with task_id = '+str(self.__task_id)


class GetTaskStatusAction(PMMCliAction):
    def __init__(self, parameter_stdin, task_param):
        PMMCliAction.__init__(self, parameter_stdin, task_param)
        self.__parameters = task_param
        self.__task_id = None

    def validate(self):
        if len(self.__parameters) != 1:
            return None, "Parameter 'task_id' not specified"
        self.__task_id = self.__parameters[0]
        # there is no format requirements for task-id
        return 1, ''

    def doActivity(self):
        if not self.__task_id:
            self.__task_id = self.__parameters[0]
        status = pmm_task.getPMMTaskManager().getTaskStatus(self.__task_id)
        if status:
            data = Data.factory( task_status = status )
            return data, 0, None
        else:
            return None, 2, 'No task found with task_id = '+str(self.__task_id)


class GetTaskLogAction(PMMCliAction):
    def __init__(self, parameter_stdin, task_param):
        PMMCliAction.__init__(self, parameter_stdin, task_param)
        self.__parameters = task_param
        self.__task_id = None

    def validate(self):
        if len(self.__parameters) != 1:
            return None, "Parameter 'task_id' not specified"
        self.__task_id = self.__parameters[0]
        # there is no format requirements for task-id
        return 1, ''

    def doActivity(self):
        if not self.__task_id:
            self.__task_id = self.__parameters[0]
        tasklog =  pmm_task.getPMMTaskManager().getTaskLog(self.__task_id)
        if tasklog:
            data = Data.factory(task_log = tasklog)
            return data, 0, None
        else:
            return None, 2, 'No task found with task_id = '+str(self.__task_id)


class RemoveTaskDataAction(PMMCliAction):
    def __init__(self, parameter_stdin, task_param):
        PMMCliAction.__init__(self, parameter_stdin, task_param)
        self.__parameters = task_param
        self.__task_id = None

    def validate(self):
        if len(self.__parameters) != 1:
            return None, "Parameter 'task_id' not specified"
        self.__task_id = self.__parameters[0]
        # there is no format requirements for task-id
        return 1, ''

    def doActivity(self):
        if not self.__task_id:
            self.__task_id = self.__parameters[0]
        result = pmm_task.getPMMTaskManager().removeTask(self.__task_id)
        if result:
            errcode = 0
            return None, errcode, None
        else:
            return None, 2, 'No task found with task_id = '+str(self.__task_id)


class StopTaskAction(PMMCliAction):
    def __init__(self, parameter_stdin, task_param):
        PMMCliAction.__init__(self, parameter_stdin, task_param)
        self.__parameters = task_param
        self.__task_id = None

    def validate(self):
        if len(self.__parameters) != 1:
            return None, "Parameter 'task_id' not specified"
        self.__task_id = self.__parameters[0]
        # there is no format requirements for task-id
        return 1, ''

    def doActivity(self):
        if not self.__task_id:
            self.__task_id = self.__parameters[0]
        result = pmm_task.getPMMTaskManager().stopTask(self.__task_id)
        errcode = 0
        if result:
            errcode = 0
            return None, errcode, None
        else:
            return None, 2, 'No task found with task_id = '+str(self.__task_id)


def usage():
    print "pmmcli. Plesk Migration Manager."
    print "(c) Parallels"
    print "" 
    print "Usage:"
    print "" 
    print "    pmmcli <action> [<param1-value>[<param2-value>[<param3-value>[<param4-value>]]]]"
    print "" 
    print "        Available action:"
    print "            --get-dumps-list"
    print "                 STDIN is 'dump-list-query' from pmm_api_xml_protocols.xsd"
    print "            --get-dump-overview"
    print "                 STDIN is 'dump-specification' from pmm_api_xml_protocols.xsd or session-id"
    print "            --check-dump"
    print "                 STDIN is 'dump-specification' from pmm_api_xml_protocols.xsd"
    print "            --export-dump-as-file"
    print "                 STDIN is 'src-dst-files-specification' from pmm_api_xml_protocols.xsd"
    print "            --export-file-as-file"
    print "                 STDIN is 'src-dst-files-specification' from pmm_api_xml_protocols.xsd"
    print "            --import-file-as-dump"
    print "                 STDIN is 'src-dst-files-specification' from pmm_api_xml_protocols.xsd"
    print "            --restore"
    print "                 STDIN is 'restore-task-description' from pmm_api_xml_protocols.xsd"
    print "            --resolve-conflicts"
    print "                 STDIN is 'resolve-conflilcts-task-description' from pmm_api_xml_protocols.xsd"
    print "            --get-conflicts-description"
    print "                 param1-value is session-id"
    print "            --get-tasks-list"
    print "                 param1-value is task type ( 'Restore' or 'Backup' )"
    print "                 param2-value is guid of the task owner (or 'any')"
    print "                 param3-value (ignored for 'Restore' task type) is id of the top object to backup ('backup-task-description/backup-specification/object-to-backup[0]/@id') or 'any' string"
    print "                 param4-value (ignored for 'Restore' task type) is type of the top object to backup ('backup-task-description/backup-specification/object-to-backup[0]/@type')"
    print "            --get-task"
    print "                 param1-value is task-id"
    print "            --make-dump"
    print "                 STDIN is 'backup-task-description' from pmm_api_xml_protocols.xsd"
    print "            --delete-dump"
    print "                 STDIN is 'dump-specification' from pmm_api_xml_protocols.xsd"
    print "            --get-task-status"
    print "                 param1-value is task-id"
    print "            --get-task-log"
    print "                 param1-value is task-id"
    print "            --remove-task-data"
    print "                 param1-value is task-id"
    print "            --stop-task"
    print "                 param1-value is task-id"
    print "            --help"
    print "                 Display this help"
    print ""
    print "    The pmm_api_xml_protocols.xsd is placed at " + pmm_config.pmm_api_xml_protocols_schema()         
    print "    The output is 'responce' from pmm_api_xml_protocols.xsd"         
    print "    The task-id, session-id is result of previous long operation."
    print ""

    
def bind_stdin(command):
    path = pmm_config.pmmcli_dir() + "/" + command[2:] + ".stdin"
    if os.path.exists(path) and not os.path.isdir(path):
        sys.stdin = open(path,"rt")

def reset_stdin():
    if sys.stdin != sys.__stdin__:
        sys.stdin = sys.__stdin__
    
def safe_print(packet):
    try:
        sys.stdout.write(packet)
    except IOError:
        pass

def validate_path(path):
    if not os.path.isabs(path):
        return (False, "Session path must be absolute: " +  path + "\n")

    if os.path.exists(path) and not os.path.isdir(path):
        return (False, "Path " + path + " exists, but not a directory.\n")
    if not os.path.exists(path):
        try:
            os.makedirs(path, 0700)
        except UnicodeError, e:
            return (False, "Invalid character in session directory\n")
        except OSError, e:
            return (False, "Unable to create directory " + path + ": " + e.strerror + "\n")
    return (True, None)

def get_dumps_list(parameters):
    dump_list_query = sys.stdin.read()
    return ActionRunner(GetDumpsListAction, dump_list_query, None).doActivity()

def get_dump_overview(parameters):
    dump_specification = sys.stdin.read()
    return ActionRunner(GetDumpOverviewAction, dump_specification, None).doActivity()

def check_dump(parameters):
    dump_specification = sys.stdin.read()
    return ActionRunner(CheckDumpAction, dump_specification, None).doActivity()

def export_dump_as_file(parameters):
    src_dst_files_specification = sys.stdin.read()
    return ActionRunner(ExportDumpAsFileAction, src_dst_files_specification, None).doActivity()

def export_file_as_file(parameters):
    src_dst_files_specification = sys.stdin.read()
    return ActionRunner(ExportFileAsFileAction, src_dst_files_specification, None).doActivity()

def import_file_as_dump(parameters):
    src_dst_files_specification = sys.stdin.read()
    return ActionRunner(ImportFileAsDumpAction, src_dst_files_specification, None).doActivity()

def restore(parameters):
    # <parameter1> is Plesk User Id which is a owner of a restore process
    # if <parameter2> is set to --force, PMM should not make conflicts detection but should just restore received Restore Specification
    restore_task_specification = sys.stdin.read()
    return ActionRunner(RestoreAction, restore_task_specification, parameters).doActivity()

def resolve_conflicts(parameters):
    resolve_conflicts_task_description = sys.stdin.read()
    return ActionRunner(ResolveConflictsAction, resolve_conflicts_task_description, None).doActivity()

def get_conflicts_description(parameters):
    # <parameter1> is session_id
    session_id = parameters
    return ActionRunner(GetConflictsDescriptionAction, None, session_id).doActivity()

def get_tasks_list(parameters):
    # <parameter1> is a task type (Restore|Backup)
    # <parameter2> (optional) is a guid of the task owner or "any" string
    # <parameter3> (optional) is a id of the top object or "any" string(for 'Backup' task type, ignored for 'Restore' task type)
    # <parameter4> (optional) is a type of the top object (for 'Backup' task type, ignored for 'Restore' task type)
    task_params = parameters
    return ActionRunner(GetTasksListAction, None, task_params).doActivity()

def get_task(parameters):
    task_param = parameters
    return ActionRunner(GetTaskAction, None, task_param).doActivity()

def make_dump(parameters):
    backup_task_description = sys.stdin.read()
    return ActionRunner(MakeDumpAction, backup_task_description, None).doActivity()

def delete_dump(parameters):
    dump_description = sys.stdin.read()
    return ActionRunner(DeleteDumpAction, dump_description, None).doActivity()

def get_task_status(parameters):
    task_param = parameters
    return ActionRunner(GetTaskStatusAction, None, task_param).doActivity()

def get_task_log(parameters):
    task_param = parameters
    return ActionRunner(GetTaskLogAction, None, task_param).doActivity()

def remove_task_data(parameters):
    task_param = parameters
    return ActionRunner(RemoveTaskDataAction, None, task_param).doActivity()

def stop_task(parameters):
    task_param = parameters
    return ActionRunner(StopTaskAction, None, task_param).doActivity()

def convertToXmlString(response, name):
    resp_str = cStringIO.StringIO()
    resp_encoded = EncoderFile(resp_str, "utf-8")
    resp_encoded.write('<?xml version="1.0" encoding="UTF-8"?>\n')
    response.export(resp_encoded, 0, name_ = name)
    packet = resp_str.getvalue()
    return packet

def main():
    actions = {"get-dumps-list": get_dumps_list,
               "get-dump-overview": get_dump_overview,
               "check-dump": check_dump,
               "export-dump-as-file": export_dump_as_file,
               "export-file-as-file": export_file_as_file,
               "import-file-as-dump": import_file_as_dump,
               "restore": restore,
               "resolve-conflicts": resolve_conflicts,
               "get-conflicts-description": get_conflicts_description,
               "get-tasks-list": get_tasks_list,
               "get-task": get_task,
               "make-dump": make_dump,
               "get-task-status": get_task_status,
               "get-task-log": get_task_log,
               "remove-task-data": remove_task_data,
               "stop-task": stop_task,
               "delete-dump": delete_dump}
    
    debug_mode = os.environ.has_key("DEBUG_PMMCLI")
    
    if len(sys.argv) < 2:
        usage()
        sys.exit(1)
    
    if sys.argv[1] == '--help':
        usage()
        sys.exit(0)
    
    if not sys.argv[1][2:] in actions:
        print "Unknown command \'" + sys.argv[1] + "'.\n"
        usage()
        sys.exit(1)
    
    pmmcli_log_dir = pmm_config.pmm_logs_directory()
    parameters = sys.argv[2:]
    validate_path(pmmcli_log_dir)
    log.initPidLog("pmmcli", pmmcli_log_dir, 1)
    
    tempstdout = cStringIO.StringIO()
    tempstderr = cStringIO.StringIO()
    
    try:
        try:
            error_message = ""
            sys.stdout = tempstdout
            sys.stderr = tempstderr
            bind_stdin(sys.argv[1])
            try:
                data_action_response, errcode_response, error_message = actions.get(sys.argv[1][2:])(parameters)
                if not errcode_response:
                    errcode_response = 0
            except PmmcliActionParamException, ex:
                errcode_response = 1
                data_action_response = None
                error_message = "Invalid input parameters in '" + sys.argv[1] + "' command:\n" + ex.get_message()
            if not error_message:
                error_message = ""
            if tempstdout.getvalue():
                error_message = error_message + "STDOUT:\n-----------------------\n" + tempstdout.getvalue() + "\n" 
            if tempstderr.getvalue():
                error_message = error_message + "STDERR:\n-----------------------\n" + tempstderr.getvalue() + "\n"
        finally:
            sys.stdout = sys.__stdout__
            sys.stderr = sys.__stderr__
            reset_stdin()
        response = Response( errcode = errcode_response, data = data_action_response, errmsg = error_message)
        packet = convertToXmlString(response,'response')
        _logger.info("Outgoing packet:\n" + packet)
        safe_print(packet)
    except pmmcli_session.RSessionException, e:
        response = Response( errcode = 3, errmsg = "Session with id='" + str(e.get_session_id()) + "' not found")
        packet = convertToXmlString(response,'response')
        _logger.info("Outgoing packet:\n" + packet)
        safe_print(packet)
    except (pmm_dump_formatter.DumpsStorageCredentialsFormatException, pmm_dump_formatter.DumpSpecificationFormatException), e:
        response = Response( errcode = 1, errmsg = "Wrong format for dump specification: " + e.get_message()+ "\n" +  stacktrace.stacktrace())
        packet = convertToXmlString(response,'response')
        _logger.info("Outgoing packet:\n" + packet)
        safe_print(packet)
        sys.exit(1)
    except PMMUtilityException, e: 
        _logger.critical("PMMUtility exception: \n" +  str(unicode(e)) + "\n" +  stacktrace.stacktrace())
        response = Response( errcode = 1000, errmsg = "pmm utility '" + e.utility() + "' raised an exception. Error code is: " + str(e.exitcode) + "\n" + "See pmmcli.log to find out detailed information on this")
        packet = convertToXmlString(response,'response')
        _logger.info("Outgoing packet:\n" + packet)
        safe_print(packet)
        sys.exit(1)
    except Exception, e:
        _logger.critical("Runtime error in pmmcli: \n" + str(e.__class__) + " " + str(e) + "\n" +  stacktrace.stacktrace())
        response = Response( errcode = 1001, errmsg = "Runtime error in pmmcli: \n" + str(e.__class__) + " " + str(e) + "\n" +  stacktrace.stacktrace())
        packet = convertToXmlString(response,'response')
        _logger.info("Outgoing packet:\n" + packet)
        safe_print(packet)
        sys.exit(1)
    except:
        _logger.critical("Unhandled exception in pmmcli: \n" +  stacktrace.stacktrace())
        response = Response( errcode = 1002, errmsg = "Unhandled exception in pmmcli: \n" +  stacktrace.stacktrace())
        packet = convertToXmlString(response,'response')
        _logger.info("Outgoing packet:\n" + packet)
        safe_print(packet)
        sys.exit(1)
    
if __name__ == '__main__':
    main()

