import pmm_config
import pmmcli_config
import pmm_task
import pmm_suspend_handler
import pmm_repository_access_service
import osutil
import os
import sys
import time
import random
import datetime
import smtplib
import StringIO

try: import subprocess
except ImportError: import compat.subprocess as subprocess

from email.MIMEMultipart import MIMEMultipart
from email.MIMEText import MIMEText
from email.Utils import formatdate
import re
import cPickle
import stacktrace


def make_xlat(*args, **kwds):
    adict = dict(*args, **kwds)
    rx = re.compile('|'.join(map(re.escape, adict)))
    def one_xlat(match):
        return adict[match.group(0)]
    def xlat(text):
        return rx.sub(one_xlat, text)
    return xlat


actions = None


mswindows = (sys.platform == "win32")


#
# Daemon task processor class
#
class DaemonTasks:
    def __init__(self):
        self.__daemontasks = []
        self.__initialize()

    def __initialize(self):
        # get daemontasks
        initialized_daemontasks = []
        daemontasks_dir = pmm_config.daemontasks_dir()
        if os.path.isdir(daemontasks_dir):
            names = os.listdir(daemontasks_dir)
            for name in names:
                fullname = os.path.join(daemontasks_dir, name)
                if os.path.isfile(fullname):
                    daemontask = DaemonTask.load(fullname)
                    if isinstance(daemontask,DaemonTask):
                        # add daemontask to daemontask actions
                        self.__daemontasks.append(daemontask)
                    initialized_daemontasks.append(fullname)
            for filename in initialized_daemontasks:
                osutil.unlink_nothrow(filename)

    def get_daemontask_actions(self):
        return [daemontask.action() for daemontask in self.__daemontasks]

#
# Base daemon task class
# Daemon tasks are created with pmmcli then readed out with pmmcli-daemon
#
class DaemonTask:
    def __init__(self):
        pass

    def save(self):
        daemontasks_dir = pmm_config.daemontasks_dir()
        if not os.path.isdir(daemontasks_dir):
            os.mkdir(daemontasks_dir)
        daemontask_filename = os.path.join(daemontasks_dir,time.strftime('%Y%m%d%H%M%S',time.localtime()) + str(random.randint(0,1000)))
        while os.path.isfile(daemontask_filename):
            daemontask_filename = os.path.join(daemontasks_dir,time.strftime('%Y%m%d%H%M%S',time.localtime()) + str(random.randint(0,1000)))
        daemontask_file = open(daemontask_filename, 'w')
        cPickle.dump(self, daemontask_file)
        daemontask_file.close()

    def load(daemontask_filename):
        daemontask_file = open(daemontask_filename, 'r')
        daemontask = cPickle.load(daemontask_file)
        daemontask_file.close()
        return daemontask

    load = staticmethod(load)

    def action(self):
        return DaemonTaskAction()


#
# Daemon task class used with --import-file-as-dump
#
class DaemonTaskDeleteDump(DaemonTask):
    def __init__(self, dump_specification_formatter):
        self.__dump_specification_formatter = dump_specification_formatter

    def action(self):
        return DaemonTaskDeleteDumpAction(self.__dump_specification_formatter)


def getActions(log):
    global actions
    if not actions:
        actions = ActionItems(log)
    return actions


class ActionItem:
    def process(self,logger):
        logger("Action " + self.__class__.__name__ + " started")


class DaemonTaskAction(ActionItem):
    def __init__(self):
        pass

    def action(self):
        return ActionItem()


class DaemonTaskDeleteDumpAction(DaemonTaskAction):
    def __init__(self, dump_specification_formatter):
        self.__dump_specification_formatter = dump_specification_formatter

    def process(self,logger):
        ActionItem.process(self,logger)
        logger("Action param: %s:\n%s" % ('dump_specification_formatter',self.__dump_specification_formatter.buildString()))
        
        # remove dump with pmm-ras
        errcode = 0
        message = None
        access_service = pmm_repository_access_service.LocalRepositoryAccessService
        dumps_storage_credentials = self.__dump_specification_formatter.get_dumps_storage_credentials_formatter()
        errcode, message = access_service(dumps_storage_credentials).deleteDump(self.__dump_specification_formatter.buildXml())
        logger("Action ended with errcode=%d" % errcode)
        if message:
            logger("Message is: %s" % message)


class CleanTaskSessionAction(ActionItem):
    def __init__(self, task_id):
        self.__task_id = task_id

    def process(self,logger):
        ActionItem.process(self,logger)
        logger("Action param: %s:%s" % ('task_id',self.__task_id))
        task_manager = pmm_task.getPMMTaskManager()
        task_manager.removeTask(self.__task_id)


class CleanOrphanedSessionAction(ActionItem):
    def __init__(self, session_path):
        self.__session_path = session_path

    def process(self,logger):
        ActionItem.process(self,logger)
        logger("Action param: %s:%s" % ('session_path',self.__session_path))
        osutil.unlink_recursively(self.__session_path)


mailer = None


class Mailer:
    def __init__(self):
        #get init params from pmmcli_config file
        self.__smtpserver = pmmcli_config.get().get_smtpserver()
        self.__authrequired = pmmcli_config.get().get_authrequired()
        self.__smtpuser = pmmcli_config.get().get_smtpuser()
        self.__smtppass = pmmcli_config.get().get_smtppass()
        self.__subject = pmmcli_config.get().get_mailsubject()
        self.__from = pmmcli_config.get().get_mailfrom()
        self.__reply_to = pmmcli_config.get().get_mailreplyto()
        self.__body = pmmcli_config.get().get_mailbody()

    def send(self,logger,mail_to,task_type,task_status,owner_guid,owner_type,creation_date,fullname,logfile):
        # replace tokens with actual values
        status = "working"
        if task_status.get_stopped():
            status = 'stopped'
        elif task_status.get_finished():
            status = task_status.get_finished().get_status()
        
        adict = {"%TASK_TYPE%" : task_type, "%TASK_STATUS%" : status, "%OWNER_GUID%" : owner_guid, "%OWNER_TYPE%" : owner_type, "%CREATION_DATE%" : creation_date, "%FULL_NAME%" : fullname }
        translate = make_xlat(adict)
        
        
        # prepare message body
        self.__body = self.__body.replace('\\n','\n').lstrip('\'"').rstrip('\'"')
        
        if self.__from == '':
            if self.__authrequired == 0:
                self.__from = "PMMCli-Daemon<%s>" % mail_to
            else:
                self.__from = self.__smtpuser
        
        msg = MIMEMultipart()
        msg['From'] = self.__from
        msg['To'] = mail_to
        msg['Date'] = formatdate(localtime=True)
        msg['Subject'] = self.__subject
        
        msg.attach( MIMEText(translate(self.__body)) )
        
        if os.path.isfile(logfile):
            part = MIMEText('Task log is attached below')
            part.set_payload(open(logfile,'rb').read())
            part.add_header('Content-Disposition','attachment; filename="%s"' % os.path.split(logfile)[1])
            msg.attach(part)
        
        if mswindows or self.__authrequired == 1:
            smtp_server_session = smtplib.SMTP(self.__smtpserver)
            try:
                if self.__authrequired:
                    smtp_server_session.login(self.__smtpuser,self.__smtppass)
            
                smtp_result = smtp_server_session.sendmail(self.__from,mail_to,msg.as_string())
                if not smtp_result:
                    logger("E-mail sent to: " + mail_to)
                else:
                    for recip in smtp_result.keys():
                        errstr = """Could not delivery mail to: %s\nServer response is: %s\n%s""" % (recip, smtp_result[recip][0], smtp_result[recip][1])
                        logger(errstr)
            
            except smtplib.SMTPException, ex:
                errstr = """Could not delivery mail to: %s\nException %s raised : %s""" % (mail_to, repr(ex), str(ex))
                logger(errstr)
                pass
        else:
            # could not use smtplib interface on linux without authorization
            # use sendmail interface
            
            mail_sender = "-fPMMCli-Daemon<%s>" % mail_to
            
            sendmail = subprocess.Popen(["sendmail",mail_sender,"-t","-oi"], stdin = subprocess.PIPE)
            sendmail_stdout = sendmail.communicate(input = msg.as_string())[0]
            logger("email sent: %s" % msg.as_string())
            
            try:
                os.waitpid(sendmail.pid,0)
            except OSError:
                pass
            logger("sendmail output: %s" % sendmail_stdout)

    def get():
        global mailer
        if not mailer:
            mailer = Mailer()
        return mailer
    
    get = staticmethod(get)


class MailToAction(ActionItem):
    def __init__(self, task_id, task_type, task_status, owner_guid, owner_type, creation_date, fullname, logfilename, mail_to):
        self.__task_id = task_id
        self.__task_type = task_type
        self.__task_status = task_status
        self.__owner_guid = owner_guid
        self.__owner_type = owner_type
        self.__creation_date = creation_date
        self.__fullname = fullname
        self.__logfilename = logfilename
        self.__mail_to = mail_to
        self.__AUTHREQUIRED = 0
        self.__smtpserver = 'localhost'
        self.__smtpuser = ''
        self.__smtppass = ''

    def process(self,logger):
        ActionItem.process(self,logger)
        logger("Action params: %s:%s, %s:%s, %s:%s, %s:%s, %s:%s, %s:%s, %s:%s" % ('task_type',self.__task_type,'owner_guid',self.__owner_guid,'owner_type',self.__owner_type,'creation_date',self.__creation_date,'fullname',self.__fullname,'logfilename',self.__logfilename,'mail_to',self.__mail_to))
        mailsender = Mailer().get()
        mailsender.send(logger,self.__mail_to,self.__task_type,self.__task_status,self.__owner_guid,self.__owner_type,self.__creation_date,self.__fullname,self.__logfilename)
        # mark task as 'mailsent'
        storage = pmm_task.PMMTaskPersistentStorage()
        task = storage.getPmmTask(str(self.__task_id))
        if task:
            task.set_mailsent()
            pmm_task.PMMTaskManager().updateTask(task)


class UnsuspendTaskAction(ActionItem):
    def __init__(self, session_id):
        self.__session_id = session_id

    def process(self,logger):
        ActionItem.process(self,logger)
        logger("Action param: %s:%s" % ('session_id',self.__session_id))
        pmm_suspend_handler.SuspendHandler().run(self.__session_id, logger)


class PmmcliLogRotateAction(ActionItem):
    def __init__(self, file_name):
        self.__file_name = file_name
        self.__max_files = pmmcli_config.get().get_max_log_files()
        self.__max_size = pmmcli_config.get().get_max_log_size()

    def __rotate(self,logger):
        startswith = self.__file_name
        startswith_len = len(self.__file_name)
        files = {}
        for filename in os.listdir(pmm_config.pmm_logs_directory()):
            if os.path.isfile(filename):
                if filename.startswith(startswith):
                    time_stamp = filename[startswith_len+1:]
                    if time_stamp != '':
                        creation_date = time.strptime(time_stamp,'%Y-%m-%d-%H%M%S')
                        creation_datetime = datetime.datetime(creation_date[0],creation_date[1],creation_date[2],creation_date[3],creation_date[4],creation_date[5])
                        files[datetime.datetime.now() - creation_datetime] = filename
        keys = files.keys( )
        keys.sort( )
        files_to_delete_count = len(keys) - self.__max_files + 1
        if files_to_delete_count > 0:
            files_to_delete = [files[key] for key in keys[-files_to_delete_count:]]
            for file_to_delete in files_to_delete:
                logger("delete old log file " + file_to_delete)
                osutil.unlink_nothrow(os.path.join(pmm_config.pmm_logs_directory(),file_to_delete))
        time_stamp = time.strftime('%Y-%m-%d-%H%M%S',time.localtime())
        logger("rename log file to " + os.path.join(pmm_config.pmm_logs_directory(),self.__file_name) + "." +time_stamp)
        os.rename(os.path.join(pmm_config.pmm_logs_directory(),self.__file_name),os.path.join(pmm_config.pmm_logs_directory(),self.__file_name) + "." + time_stamp)

    def process(self,logger):
        ActionItem.process(self,logger)
        logger("Action param: %s:%s" % ('file_name',self.__file_name))
        rotate = False
        fullname = os.path.join(pmm_config.pmm_logs_directory(),self.__file_name)
        if os.path.isfile(fullname) and os.path.getsize(fullname) >= self.__max_size:
            rotate = True
        if rotate:
            self.__rotate(logger)


class SessionCleaner:
    def __init__(self):
        self.__task_sessions = []
        self.__days_to_keep_sessions = pmmcli_config.get().get_days_to_keep_sessions()
        self.__initialize()

    def __initialize(self):
        task_manager = pmm_task.getPMMTaskManager()
        backup_task_list = task_manager.realGetTaskList('Backup')
        for task in backup_task_list:
            time_corrupted = False
            try:
                creation_datetime = datetime.datetime.fromtimestamp(task.get_creation_date())
            except ValueError:
                # task creation date is corrupted - add task to the task list to clear
                time_corrupted = True
            if time_corrupted:
                self.__task_sessions.append(task.get_pmm_pid())
            else:
                delta = datetime.datetime.now() - creation_datetime
                if delta.days >= self.__days_to_keep_sessions:
                    self.__task_sessions.append(task.get_pmm_pid())

    def get_task_sessions(self):
        return self.__task_sessions


class RestoreSessionCleaner:
    def __init__(self):
        self.__task_sessions = []
        self.__orphaned_sessions = []
        self.__days_to_keep_sessions = pmmcli_config.get().get_days_to_keep_sessions()
        self.__initialize()

    def __initialize(self):
        session_paths = []
        # fill in restore sessions acessible with pmm tasks
        task_manager = pmm_task.getPMMTaskManager()
        restore_task_list = task_manager.realGetTaskList('Restore')
        for task in restore_task_list:
            session_paths.append(task.get_session_path())
            time_corrupted = False
            try:
                creation_datetime = datetime.datetime.fromtimestamp(task.get_creation_date())
            except ValueError:
                # task creation date is corrupted - add task to the task list to clear
                time_corrupted = True
            if time_corrupted:
                self.__task_sessions.append(task.get_pmm_pid())
            else:
                delta = datetime.datetime.now() - creation_datetime
                if delta.days >= self.__days_to_keep_sessions:
                    self.__task_sessions.append(task.get_pmm_pid())
        
        # find orphaned restore sessions
        restore_sessions_directory = pmm_config.restore_session_dir()
        if os.path.isdir(restore_sessions_directory):
            names = os.listdir(restore_sessions_directory)
            for name in names:
                fullname = os.path.join(restore_sessions_directory, name)
                if os.path.isdir(fullname):
                    if fullname not in session_paths:
                        # this is not finished restore session
                        mod_time = datetime.datetime.fromtimestamp(os.path.getmtime(fullname))
                        delta = datetime.datetime.now() - mod_time
                        if delta.days >= self.__days_to_keep_sessions:
                            self.__orphaned_sessions.append(fullname)


    def get_task_sessions(self):
        return self.__task_sessions

    def get_orphaned_sessions(self):
        return self.__orphaned_sessions


class ActionItems:
    def __init__(self, log):
        self.__logger = log
        self.__action_items = []
        self.__initialize()

    def __initialize(self):
        
        # Get restore tasks list
        task_manager = pmm_task.getPMMTaskManager()
        restore_task_list = task_manager.realGetTaskList('Restore')
        for task in restore_task_list:
            task_status = task_manager.operatorGetTaskStatus(task)
            working = task_status.get_working()
            if not working:
                mailto = task.get_mailto()
                if mailto and mailto != '' and not task.get_mailsent():
                    self.__logger("MailToAction added")
                    self.__action_items.append(MailToAction(task.get_pmm_pid(),'Restore',task_status,task.get_owner_guid(),task.get_owner_type(),task.get_creation_date_formatted(),task.get_fullname(),task.get_migration_result_filename(),task.get_mailto()))
                finished = task_status.get_finished()
                if finished and finished.get_status() != 'success':
                    self.__logger("UnsuspendTaskAction added")
                    self.__action_items.append(UnsuspendTaskAction(task.get_session_path()))
        
        
        # Get backup tasks list
        backup_task_list = task_manager.realGetTaskList('Backup')
        for task in backup_task_list:
            task_status = task_manager.operatorGetTaskStatus(task)
            working = task_status.get_working()
            if not working:
                mailto = task.get_mailto()
                if mailto and mailto != '' and not task.get_mailsent():
                    self.__logger("MailToAction added")
                    self.__action_items.append(MailToAction(task.get_pmm_pid(),'Backup',task_status,task.get_owner_guid(),task.get_owner_type(),task.get_creation_date_formatted(),task.get_fullname(),task.get_migration_result_filename(),task.get_mailto()))
                finished = task_status.get_finished()
                if finished and finished.get_status() != 'success':
                    self.__logger("UnsuspendTaskAction added")
                    self.__action_items.append(UnsuspendTaskAction(task.get_session_path()))
                    
        # Get backup task sessions to clean
        for session in SessionCleaner().get_task_sessions():
            self.__logger("CleanTaskSessionAction added")
            self.__action_items.append(CleanTaskSessionAction(session))
        
        # Get restore task sessions to clean
        restoreSessionCleaner = RestoreSessionCleaner()
        for session in restoreSessionCleaner.get_task_sessions():
            self.__logger("CleanTaskSessionAction added")
            self.__action_items.append(CleanTaskSessionAction(session))
        
        # Get orphaned restore sessions to clean
        for session in restoreSessionCleaner.get_orphaned_sessions():
            self.__logger("CleanOrphanedSessionAction added")
            self.__action_items.append(CleanOrphanedSessionAction(session))

        self.__logger("PmmcliLogRotateAction added")
        self.__action_items.append(PmmcliLogRotateAction('pmmcli.log'))

        self.__logger("PmmcliLogRotateAction added")
        self.__action_items.append(PmmcliLogRotateAction('pmmcli_daemon.log'))

        self.__logger("PmmcliLogRotateAction added")
        self.__action_items.append(PmmcliLogRotateAction('migration.log'))

    def get_action_items(self):
        return self.__action_items

    def reason(self):
        reason = False
        # Get restore tasks list
        task_manager = pmm_task.getPMMTaskManager()
        restore_task_list = task_manager.realGetTaskList(['Backup','Restore'])
        for task in restore_task_list:
            task_status = task_manager.operatorGetTaskStatus(task)
            working = task_status.get_working()
            if working:
                reason = True
                break
        return reason

    def process(self, logger):
        for action_item in self.get_action_items():
            try:
                action_item.process(logger)
            except Exception, e:
                logger("exception during action processing\nStacktrace is: " + stacktrace.stacktrace())

    def process_daemontasks(self,logger):
        daemon_tasks = DaemonTasks()
        for action_item in daemon_tasks.get_daemontask_actions():
            try:
                action_item.process(logger)
            except Exception, e:
                logger("exception during action processing\nStacktrace is: " + stacktrace.stacktrace())
